import pathlib
import curses
import os
import shutil
import threading
import subprocess
import time

from cursedspace import Key, Panel, colors, ShellContext

from metaindexmanager import command
from metaindexmanager import utils
from metaindexmanager.utils import logger
from metaindexmanager.panel import ListPanel


class Line:
    def __init__(self, panel, path, stat=None):
        self.panel = panel
        self.path = path
        self.is_dir = path.is_dir()
        self.is_symlink = path.is_symlink()
        if stat is not None:
            self.stat = stat
        else:
            try:
                self.stat = path.stat()
            except:
                self.stat = None
        self.name = path.name
        self.stem = path.stem
        self.suffix = path.suffix
        self.fullpath = path
        self.icon = utils.get_ls_icon(path, self.stat)
        self.attr, fg, bg = utils.get_ls_colors(path, self.stat)

        self.fg = colors.DEFAULT.foreground if fg is None else fg
        self.bg = -1 if bg is None else bg

    def copy(self):
        return Line(self.panel, self.path, self.stat)

    def __eq__(self, other):
        return other is self \
            or (isinstance(other, pathlib.Path) and other == self.path) \
            or (isinstance(other, str) and other == str(self.path)) \
            or (isinstance(other, Line) and other.path == self.path)

    def __lt__(self, other):
        return self.name < other.name

    def __str__(self):
        prefix = "d" if self.is_dir else "f"
        return f"{prefix}{self.name}"


class FolderObserver(threading.Thread):
    def __init__(self, panel, path, items=None):
        super().__init__()
        self.path = path
        self.panel = panel
        self.items = items[:] if items else []
        self.active = False
        self.do_interrupt = threading.Event()
    
    def run(self):
        self.active = True
        text = None
        while self.active:
            items = [item for item in self.path.iterdir()]
            newtext = "\n".join([str(i) for i in sorted(items)])
            if text is None or text != newtext:
                text = newtext
                self.panel.app.callbacks.put((self.panel, lambda: self.on_change(items)))
            self.do_interrupt.wait(2)  # TODO: should be configurable

    def on_change(self, items):
        if self.active:
            return self.panel.finish_change_path([Line(self.panel, i) for i in items])
        return False

    def close(self):
        self.active = False
        self.do_interrupt.set()
        self.join()


class FilePanel(ListPanel):
    SCOPE = 'files'

    CONFIG_ICONS = 'use-icons'

    def __init__(self, *args, path='.', **kwargs):
        super().__init__(*args, **kwargs)
        self.path = None
        self.previous_path = None
        self.post_load = None
        self.show_hidden_files = False
        self.fancy_icons = False
        self.observer = None
        self.cursor_cache = {}
        self.configuration_changed()
        self.change_path(path)

    def configuration_changed(self, name=None):
        super().configuration_changed(name)

        changed = False

        if name is None or name == self.CONFIG_ICONS:
            value = self.app.configuration.bool(self.SCOPE, self.CONFIG_ICONS, 'no')
            changed = self.fancy_icons != value
            self.fancy_icons = value

        if changed:
            if self.win is not None:
                self.scroll()
                self.paint(True)

    def change_path(self, path):
        logger.debug(f"FilePanel: changing to path {path}")
        if self.path is not None:
            self.cursor_cache[self.path] = self.cursor
        self.previous_path = self.path
        path = pathlib.Path(path).expanduser().resolve()
        if path != self.path and self.observer is not None:
            self.observer.active = False
            self.observer = None
        self.path = path
        self.offset = 0
        self.cursor = 0
        self.items = None
        if self.observer is None:
            self.observer = FolderObserver(self, path)
            self.observer.start()

    def finish_change_path(self, items):
        self.items = [i.copy() for i in sorted(items, key=lambda f: [not f.is_dir, f.name])
                      if self.show_hidden_files or not i.name.startswith('.')]

        if self.path in self.cursor_cache:
            self.cursor = self.cursor_cache[self.path]

        elif self.previous_path is not None and self.path == self.previous_path.parent:
            self.cursor = 0
            for nr, that in enumerate(self.items):
                if that.fullpath == self.previous_path:
                    self.cursor = nr
                    break

        self.scroll()
        # drop elements from the multi selection that are no longer there
        self.multi_selection = [item for item in self.multi_selection if item in self.items]
        # trigger repaint of focus bar to update the location
        self.app.callbacks.put((self, lambda: self.app.paint_focus_bar()))

        if self.post_load is not None:
            self.post_load()
            self.post_load = None

        return True

    def title(self):
        return str(self.path)

    @property
    def selected_path(self):
        return self.selected_item.fullpath

    @property
    def selected_paths(self):
        return [i.fullpath for i in self.selected_items]

    def open_selected(self):
        item = self.selected_item

        if item is None:
            self.app.error("Nothing selected")
            return

        if item.is_dir:
            self.change_path(item.fullpath)
            self.paint(True)
        else:
            self.app.open_file(item.fullpath)

    def mkdir(self, name):
        if len(name) == '':
            self.app.error("No folder name given")
            return

        new_path = self.path / name

        try:
            new_path.mkdir(parents=True, exist_ok=True)
        except Exception as exc:
            self.app.error(f"Failed to create directory: {exc}")

        self.reload()

    def delete(self, items):
        self.app.confirm(f"Are you sure you want to delete '{', '.join([item.name for item in items])}'?",
                         'n',
                         lambda: self.run_blocking(self.do_delete, items))

    def do_delete(self, context, items):
        for nr, item in enumerate(items):
            if len(items) > 1:
                context.progress((nr, len(items)))
            if item.is_dir():
                try:
                    shutil.rmtree(item)
                except Exception as exc:
                    self.app.error(f"Failed to delete the folder '{item.name}': {exc}")
            elif item.exists():
                try:
                    item.unlink()
                except Exception as exc:
                    self.app.error(f"Failed to delete '{item.name}': {exc}")

    def reload(self):
        self.change_path(self.path)

    def paint(self, clear=False):
        if self.items is None:
            Panel.paint(self, clear)
            x, y = 0, 0
            if (self.border & self.BORDER_TOP) != 0:
                y += 1
            if (self.border & self.BORDER_LEFT) != 0:
                x += 1
            self.win.addstr(y, x, "...")
            self.win.noutrefresh()
        else:
            super().paint(clear)

    def do_paint_item(self, y, x, maxwidth, is_selected, item):
        attr = item.attr

        if is_selected:
            attr |= curses.A_STANDOUT

        attr |= colors.attr((item.fg, item.bg))

        icon = '' if not self.fancy_icons else item.icon + " "

        text = f"{icon}{item.name}"
        if item in self.multi_selection:
            text = "✔ " + text
            attr |= curses.A_BOLD
        textlen = len(text)
        try:
            self.win.addstr(y, x, " "*maxwidth, attr)
        except:
            pass
        self.win.addstr(y, x, text[:maxwidth], attr)

    def jump_to(self, item, path=None):
        logger.debug(f"Jump to {path}/{item}")
        if path is None:
            path = item.parent

        if path != self.path:
            self.post_load = lambda: self.do_jump_to(item)
            self.change_path(path)
        elif self.items is None:
            self.post_load = lambda: self.do_jump_to(item)
        else:
            self.do_jump_to(item)

    def do_jump_to(self, item):
        logger.debug(f"jump to {item}")
        if item in self.items:
            self.cursor = self.items.index(item)
            self.scroll()
            if not self.is_busy:
                self.paint()

    def on_focus(self):
        if pathlib.Path.cwd() != self.path:
            os.chdir(self.path)

    def on_close(self):
        logger.debug(f"Closing {self}")
        if self.observer is not None:
            self.observer.close()


@command.registered_command
class GoToParent(command.Command):
    """Go up the file hierarchy"""
    NAME = 'go-to-parent'
    ACCEPT_IN = (FilePanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        context.panel.change_path(context.panel.path.parent)
        context.panel.paint(True)


@command.registered_command
class NewFilePanel(command.Command):
    """Create a new file manager panel"""
    NAME = 'new-file-panel'

    def execute(self, context, *args):
        logger.debug(f'new file panel at {args}')
        path = '.'
        if len(args) > 0:
            path = args[0]
        panel = FilePanel(context.application, path=path)
        context.application.add_panel(panel)
        context.application.activate_panel(panel)


@command.registered_command
class MakeFolder(command.Command):
    """Create a new folder"""
    NAME = 'mkdir'
    ACCEPT_IN = (FilePanel,)

    def execute(self, context, *args):
        if context.panel.is_busy:
            return
        context.panel.mkdir(" ".join(args))


@command.registered_command
class PasteItem(command.Command):
    """Paste content of clipboard here"""
    NAME = 'paste'
    ACCEPT_IN = (FilePanel,)

    def execute(self, context, clipboard=None):
        if clipboard is None:
            clipboard = context.application.DEFAULT_CLIPBOARD
        # TODO: implement this
        context.application.error("Not implemented yet")


@command.registered_command
class ChangeDirectory(command.Command):
    """Change directory"""
    NAME = 'cd'
    ACCEPT_IN = (FilePanel,)

    def execute(self, context, target=None):
        if target is None:
            context.application.error("No path given")
            return

        if context.panel.is_busy:
            return

        if not isinstance(target, pathlib.Path):
            try:
                target = pathlib.Path(target).expanduser().resolve()
            except:
                context.application.error(f"{target} is not a valid path")
                return

        if not target.is_dir():
            context.application.error(f"{target} does not exist")
            return

        context.panel.change_path(target)
        context.panel.focus()
        context.panel.paint(True)


@command.registered_command
class ToggleShowHiddenFiles(command.Command):
    """Toggle show hidden files"""
    NAME = 'toggle-hidden'
    ACCEPT_IN = (FilePanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        context.panel.show_hidden_files = not context.panel.show_hidden_files
        context.panel.change_path(context.panel.path)
        context.panel.paint(True)


@command.simple_command('shell', accept_in=(FilePanel,))
def shell_command(context, *args):
    """Start a shell here"""
    if context.panel.is_busy:
        return

    wait_after_execution = True

    if len(args) == 0:
        shell = shutil.which(os.getenv('SHELL') or "")
        if shell is None:
            context.application.error("No SHELL found")
            return
        wait_after_execution = False
        args = [shell]

    logger.debug(f"Running in shell: {args}")
    scr = context.application.screen
    with ShellContext(scr, True):
        try:
            subprocess.run(args, shell=True, check=False)
        except (KeyboardInterrupt,):
            pass

        if wait_after_execution:
            scr.timeout(-1)
            scr.addstr("Press return to continue...")
            scr.get_wch()
            scr.timeout(context.application.key_timeout)
    context.application.paint(True)



import curses

from metaindexmanager import command
from metaindexmanager.utils import logger
from metaindexmanager.panel import ListPanel
from metaindexmanager.docpanel import DocPanel
from metaindexmanager.filepanel import FilePanel


class DetailPanel(ListPanel):
    META_TAG_ORDER = {'title': 10,
                      'series': 11,
                      'series_index': 12,
                      'author': 20,
                      'tags': 30,
                      'subject': 30,
                      'contributors': 40,
                      'language': 50,
                      'type': 60,
                      'filename': 100,
                      'mimetype': 100,
                      'last_modified': 200,
                      }

    def __init__(self, filepath, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filepath = filepath

        self.rev_synonyms = {}
        self.synonyms = self.app.metaindexconf.synonyms
        for key, names in self.synonyms.items():
            for name in names:
                if name not in self.rev_synonyms:
                    self.rev_synonyms[name] = []
                self.rev_synonyms[name].append(key)

        self.reload()

    def title(self):
        return str(self.filepath)

    @property
    def selected_path(self):
        return self.filepath

    @property
    def selected_paths(self):
        return [self.selected_path]

    def reload(self):
        self.items = []
        metadata = [entry.metadata for entry in self.app.cache.get(self.filepath) if entry.path == self.filepath]

        if len(metadata) == 0:
            return
        metadata = metadata[0]

        keys = set(metadata.keys())
        keys = list(keys | set(sum([synonyms for key, synonyms in self.rev_synonyms.items()
                                             if key in metadata.keys()], start=[])))
        keys.sort(key=lambda k: [DetailPanel.META_TAG_ORDER.get(k, 999), '.' in k, k.lower()])

        prefix = ''
        self.items.append(('General', curses.A_BOLD))
        for key in keys:
            displaykey = key
            if '.' in key:
                keyprefix, displaykey = key.split('.', 1)
                if keyprefix != prefix:
                    prefix = keyprefix
                    self.items.append(("", curses.A_NORMAL))
                    self.items.append((prefix.title(), curses.A_BOLD))
            self.items.append((displaykey.title(), curses.A_NORMAL))

            expanded = self.synonyms.get(key, [key])
            for expkey in expanded:
                for value in metadata.getall(expkey, []):
                    self.items.append((" "*3 + self.app.humanize(value), curses.A_NORMAL))

    def do_paint_item(self, y, x, maxwidth, is_selected, item):
        text, attr = item
        self.win.addstr(y, x, " "*maxwidth)
        self.win.addstr(y, x, text[:maxwidth], attr)


@command.registered_command
class ShowDetails(command.Command):
    """Show all (metadata) details of the selected file"""
    NAME = 'details'
    ACCEPT_IN = (DocPanel, FilePanel)

    def execute(self, context):
        item = context.panel.selected_path

        panel = DetailPanel(item, context.application)
        context.application.add_panel(panel)
        context.application.activate_panel(panel)




import os
import curses
import stat
import logging
import shutil
import pathlib
import subprocess

import multidict

import metaindex.shared
import metaindex.ocr
import metaindex.indexer

from cursedspace import colors

from .shared import PROGRAMNAME


class MetaindexmanagerLogger:
    def __init__(self):
        self.logger = None
        self.formatter = logging.Formatter('[%(levelname)s] %(message)s')
        self.handler = None

    def setup(self, level=logging.WARNING, filename=None):
        if filename is None:
            self.handler = logging.NullHandler()
        else:
            self.handler = logging.FileHandler(filename, encoding='utf-8')

        self.logger = logging.getLogger(PROGRAMNAME)
        self.logger.propagate = False
        self.logger.setLevel(level)
        self.logger.addHandler(self.handler)
        self.handler.setFormatter(self.formatter)

    def __getattr__(self, name):
        if hasattr(self.logger, name):
            return getattr(self.logger, name)
        raise AttributeError(name)

    def fatal(self, *args, **kwargs):
        return self.logger.fatal(*args, **kwargs)

    def error(self, *args, **kwargs):
        return self.logger.error(*args, **kwargs)

    def warning(self, *args, **kwargs):
        return self.logger.warning(*args, **kwargs)

    def info(self, *args, **kwargs):
        return self.logger.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        return self.logger.debug(*args, **kwargs)


logger = MetaindexmanagerLogger()

LS_COLORS_FALLBACK = "no=0:di=1:ex=0;31"
LS_COLORS = {}
LS_ICONS_FALLBACK = ':'.join([
    "di=", "fi=", "ln=", "so=", "ex=",
    # documents
    "*.doc=", "*.docx=", "*.odt=", "*.gdoc=", "*.rtf=",
    "*.pdf=", "*.ps=", "*.djvu=",
    # books
    "*.epub=", "*.mobi=", "*.azw=", "*.azw3=", "*.ebook=",
    # plain text like
    "*.txt=", "*.eml=", "*.markdown=", "*.md=", "*.mkd=",
    "*.conf=", "*.cnf=", "*.cfg=", "*.ini=", "*.vim=",
    "*.json=", "*.yaml=", "*.yml=",
    # source code
    "*.c=", "*.h=", "*.cpp=", "*.cxx=", "*.hpp=", "*.hxx=", "*.cs=",
    "*.css=", "*.html=", "*.htm=", "*.xhtml=",
    "*.py=", "*.pyc=", "*.pyb=",
    "*.js=",
    "*.diff=", "*.patch=",
    # shell/script files
    "*.sh=", "*.bash=", "*.fish=", "*.zsh=", "*.ksh=",
    "*.awk=", "*.bat=", "*.cmd=", "*.ps1=",
    # archives,  or 
    "*.zip=", "*.rar=", "*.gz=", "*.xz=", "*.bz2=", "*.ace=",
    "*.bz=", "*.tar=", "*.lz=", "*.lzh=", "*.lzo=", "*.lzma=",
    "*.tgz=", "*.tbz=", "*.taz=", "*.tbz2=", "*.tz=", "*.tzo=",
    "*.deb=", "*.rpm=",
    # executables / binaries
    "*.apk=",
    "*.dll=", "*.exe=", "*.msi=",
    "*.so=",
    "*.db=", "*.sql=", "*.sqlite3=",
    "*.iso=",
    # spreadsheet
    "*.xls=", "*.ods=", "*.csv=", "*.gsheet=", "*.tsv=", "*.xlsx=",
    # presentation
    "*.ppt=", "*.pptx=", "*.gslides=", "*.odp=",
    # images  or 
    "*.jpg=", "*.jpeg=", "*.jpe=", "*.png=", "*.mng=", "*.bmp=",
    "*.tif=", "*.tiff=", "*.gif=", "*.ico=", "*.image=",
    "*.jif=", "*.jfif=", "*.jfi=", "*.svg=",
    "*.avif=", "*.webp=",
    "*.ai=", "*.psd=",
    # audio  or 
    "*.mp3=", "*.ogg=", "*.wav=", "*.aac=", "*.flac=", "*.m4a=",
    # video ,  or 
    "*.avi=", "*.mpg=", "*.mp4=", "*.ogv=", "*.webm=",
    # xml-like
    "*.xml=",
    "*.rss=",
    "*.gpx=",
    # PIM
    "*.vcf=", "*.ics=", # "*.fb=",
    # Fonts
    "*.ttf=", "*.otf=", "*.eot=", "*.font=", "*.woff=", "*.woff2=",
    # Apple stuff
    "*.ds_store=", "*.localized=",
    # temp, etc
    "*.lock=",
])
LS_ICONS = {}
DEFAULT_XDG_DIR_ICONS = [
    ('templates', ""),
    ('desktop', ""),
    ('download', ""),
    ('publicshare', ""),
    ('documents', ""),
    ('music', ""),
    ('pictures', ""),
    ('videos', ""),
]
XDG_DIR_NAMES = [name for name, _ in DEFAULT_XDG_DIR_ICONS]
USERDIR_ICONS_FALLBACK = ':'.join([f"{name}={icon}" for name, icon in DEFAULT_XDG_DIR_ICONS])
USERDIR_ICONS = {}
TERM_COLOR_MAP = {
    '31': curses.COLOR_RED,
    '32': curses.COLOR_GREEN,
    '33': curses.COLOR_YELLOW,
    '34': curses.COLOR_BLUE,
    '35': curses.COLOR_MAGENTA,
    '36': curses.COLOR_CYAN,
    '37': curses.COLOR_WHITE,
    '40': curses.COLOR_BLACK,
    '41': curses.COLOR_RED,
    '42': curses.COLOR_GREEN,
    '43': curses.COLOR_YELLOW,
    '44': curses.COLOR_BLUE,
    '45': curses.COLOR_MAGENTA,
    '46': curses.COLOR_CYAN,
    '47': -1,
}


def parse_key_sequence(text):
    sequence = []

    pos = 0
    while pos < len(text):
        token = text[pos]
        if token == '<':
            other = text.find('>', pos)
            if other < 0:
                return None
            sequence.append(text[pos:other+1])
            pos = other
        elif token == '^' and pos < len(text)-1:
            pos += 1
            if text[pos] == '^':
                sequence.append('^')
            else:
                sequence.append('^' + text[pos])
        else:
            sequence.append(token)

        pos += 1

    return tuple(sequence)


def parse_ls_colors():
    global LS_COLORS
    LS_COLORS = {}

    for pair in os.getenv('LS_COLORS', LS_COLORS_FALLBACK).split(':'):
        if '=' not in pair:
            continue
        key, value = pair.split('=', 1)
        value = value.split(';')

        attr = curses.A_NORMAL
        fg = None
        bg = None

        if len(value) == 0:
            continue
        if len(value) > 0:
            if value[0] == '0':
                pass
            if value[0] == '1':
                attr = curses.A_BOLD
        if len(value) > 1:
            fg = TERM_COLOR_MAP.get(value[1], None)
        if len(value) > 2:
            bg = TERM_COLOR_MAP.get(value[2], -1)

        LS_COLORS[key] = (attr, fg, bg)


def parse_ls_icons():
    global LS_ICONS
    LS_ICONS = {}

    global USERDIR_ICONS
    USERDIR_ICONS = {}

    # get the user defined mapping of dir [type] -> icon from $USERDIR_ICONS
    dirname_icons = []
    xdguserdir = shutil.which('xdg-user-dir')
    home = pathlib.Path().home()
    for pair in os.getenv('USERDIR_ICONS', USERDIR_ICONS_FALLBACK).split(':'):
        if '=' not in pair:
            continue

        key, icon = pair.split('=', 1)
        if len(icon) != 1:
            continue
        path = None
        if key.lower() in XDG_DIR_NAMES and xdguserdir is not None:
            result = subprocess.run([xdguserdir, key.upper()], capture_output=True)
            if result is None:
                xdguserdir = None
            else:
                path = pathlib.Path(str(result.stdout, 'utf-8').strip())
                if path == home:
                    path = None

        if path is None:
            path = pathlib.Path(key).expanduser()
        USERDIR_ICONS[path] = icon

    for pair in os.getenv('LS_ICONS', LS_ICONS_FALLBACK).split(':'):
        if '=' not in pair:
            continue

        key, icon = pair.split('=', 1)
        if len(icon) != 1:
            continue

        LS_ICONS[key] = icon


def get_ls_colors(path, stats=None):
    """Get the term attributes and colors for node at path

    Returns a tuple (attribute, foreground color, background color) based on LS_COLORS

    Foreground and background color may be None if not specified, otherwise some color from cursedspace.colors.
    The attribute will be at the very least curses.A_NORMAL.

    You should call parse_ls_colors prior to calling this to see any results.
    """
    tests = []

    # LS_COLORS:
    # di: directory
    # ex: executable
    # fi: regular file
    # pi: named pipe
    # so: socket
    # bd: block device
    # cd: character device
    # ln: symlink
    # or: symlink without target
    if path.is_dir():
        tests += ['di']
    if path.is_symlink():
        tests += ['ln']
    if path.is_socket():
        tests += ['so']
    if path.is_block_device():
        tests += ['bd']
    if path.is_char_device():
        tests += ['cd']
    if path.is_fifo():
        tests += ['pi']
    if path.is_file():
        if stats is None:
            try:
                stats = path.stat()
            except:
                stats = None
        if hasattr(stats, "st_mode") and (stats.st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)) != 0:
            tests += ['ex']
        else:
            tests += ['fi']
        if len(path.suffix) > 0:
            tests = ['*' + path.suffix] + tests
    tests.append('no')
    
    attr = curses.A_NORMAL
    fg = None
    bg = None
    for test in reversed(tests):
        if test not in LS_COLORS:
            continue
        attr, ifg, ibg = LS_COLORS[test]
        fg = fg if ifg is None else ifg
        bg = bg if ibg is None else ibg
    return (attr, fg, bg)


def get_ls_icon(path, stats=None):
    """Get the NerdFont/FontAwesome unicode icon for the node at path

    Returns the icon or a " " (space) character based on LS_ICONS.

    You should call parse_ls_icons prior to calling this to see any results.
    """
    global LS_ICONS

    if path.is_dir() and path in USERDIR_ICONS:
        return USERDIR_ICONS[path]

    tests = ['no']
    if path.is_symlink():
        tests += ['ln']
    if path.is_dir():
        tests += ['di']
    if path.is_file():
        if stats is None:
            try:
                stats = path.stat()
            except:
                stats = None
        if hasattr(stats, "st_mode") and (stats.st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)) != 0:
            tests += ['ex']
        else:
            tests += ['fi']
        if len(path.suffix) > 0:
            tests += ['*' + path.suffix.lower()]

    for test in reversed(tests):
        if test not in LS_ICONS:
            continue
        return LS_ICONS[test]
    return " "


def collection_meta_as_writable(data, basepath):
    """Change the 'data', as read from stores.get_for_collection(), to be usable for stores.store

    basepath is the parent path of the metadata file.
    """
    # turn the collection_extra again into the editable plain form,
    # i.e. restore '*', '**', and filename entries
    for key in set(data.keys()):
        if key == basepath:
            nkey = '*'
            if data[key].popall(metaindex.shared.IS_RECURSIVE) == [True]:
                nkey = '**'
            data[nkey] = data.pop(key)
            key = nkey
        else:
            nkey = key.name
            if isinstance(data, multidict.MultiDict):
                specifics = data.popall(key, [])
                if len(specifics) > 1:
                    merged = multidict.MultiDict()
                    for specific in specifics:
                        merged.extend(specific)
                    specifics = merged
            else:
                assert isinstance(data, dict)
                specifics = data.pop(key)

            data[nkey] = specifics
            key = nkey
        data[key].popall(metaindex.shared.IS_RECURSIVE, [])
    return data


def first_line(text):
    """Return only the first line of a potential multi-line text"""
    return text.split("\n")[0].split("\r")[0]


def do_ocr(path, **kwargs):
    ocr = metaindex.ocr.TesseractOCR(True, **kwargs)
    results = metaindex.indexer.index_files([path], 1, ocr, True)

    if len(results) == 0:
        return False, None

    _, success, extra = results[0]
    fulltext = extra.get('ocr.fulltext', '')
    return success and len(fulltext) > 0, fulltext




from cursedspace import Panel


class KeyHelpPanel(Panel):
    def __init__(self, candidates, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border = Panel.BORDER_ALL
        self.candidates = candidates
        self.padding = 5

    def autosize(self):
        height, width = self.app.size()
        # len(...)+2 is the space reserved for the borders
        self.resize(min(height-2, len(self.candidates)+2), width)
        self.move(height-self.dim[0]-1, 0)

    def paint(self):
        super().paint()
        if len(self.candidates) == 0:
            return

        column_widths = [max([len(line[column]) for line in self.candidates])
                         for column in range(len(self.candidates[0]))]
        # the last column stretches at most over the remaining length
        column_widths[-1] = (self.dim[1] - 2 - self.padding*(len(self.candidates[0])-1)) - sum(column_widths[:-1])
        for y, line in enumerate(sorted(self.candidates)):
            x = 1
            for column in range(len(column_widths)):
                if column_widths[column] <= 0 or x >= self.dim[1] - 2:
                    return
                self.win.addstr(y+1, x, line[column][:column_widths[column]])
                x += self.padding + column_widths[column]
        self.win.noutrefresh()




import pathlib
import tempfile
import subprocess
import string

import multidict

from cursedspace import ShellContext, InputLine

import metaindex.shared
import metaindex.indexer
import metaindex.indexers
import metaindex.ocr

from metaindexmanager import layouts
from metaindexmanager import utils
from metaindexmanager.command import registered_command, Command, simple_command
from metaindexmanager.docpanel import DocPanel
from metaindexmanager.filepanel import FilePanel
from metaindexmanager.detailpanel import DetailPanel
from metaindexmanager.keyhelppanel import KeyHelpPanel
from metaindexmanager.editorpanel import EditorPanel
from metaindexmanager.utils import parse_key_sequence, logger


@registered_command
class QuitCommand(Command):
    """Quit metaindexmanager"""
    NAME = 'quit'

    def execute(self, context):
        for panel in context.application.panels:
            panel.on_close()
        context.application.panels = []


@registered_command
class ClosePanel(Command):
    """Close the current panel"""
    NAME = 'close'

    def execute(self, context):
        context.application.close_panel(context.panel)


@registered_command
class OpenItem(Command):
    """Open the selected file or folder"""
    NAME = 'open'
    ACCEPT_IN = (DocPanel, FilePanel, EditorPanel)

    def execute(self, context):
        context.panel.open_selected()


@registered_command
class FocusNextPanel(Command):
    """Put the focus on the next panel"""
    NAME = 'next-panel'

    def execute(self, context):
        if context.panel not in context.application.panels:
            return
        idx = context.application.panels.index(context.panel)
        next_panel = context.application.panels[(idx + 1) % len(context.application.panels)]
        context.application.activate_panel(next_panel)


@registered_command
class FocusPreviousPanel(Command):
    """Put the focus on the next panel"""
    NAME = 'previous-panel'

    def execute(self, context):
        if context.panel not in context.application.panels:
            return
        idx = context.application.panels.index(context.panel)
        next_panel = context.application.panels[(idx - 1) % len(context.application.panels)]
        context.application.activate_panel(next_panel)


class PanelNumberReader(InputLine):
    def __init__(self, application):
        height, width = application.size()
        super().__init__(application, width, (height-1, 0), prefix="Focus panel: ")

    def handle_key(self, key):
        self.app.activate_panel(self.app.previous_focus)
        self.destroy()
        if len(str(key)) == 1 and str(key) in string.digits:
            FocusPanel.do_focus_panel(self.app, int(str(key)))
        self.app.hide_key_help()
        self.app.paint()


@registered_command
class FocusPanel(Command):
    """Focus the named panel"""
    NAME = 'focus'

    def execute(self, context, panel=None):
        if panel is None:
            lines = [(str(idx+1), panel.title())
                     for idx, panel in enumerate(context.application.panels)]
            panel = KeyHelpPanel(lines, context.application)
            context.application.key_help_panel = panel
            panel.autosize()
            reader = PanelNumberReader(context.application)
            context.application.activate_panel(reader)
            context.application.paint()
            return

        try:
            panel = int(panel)
        except:
            context.application.error(f"Not a number: {panel}")
            return

    @staticmethod
    def do_focus_panel(application, panelnr):
        # panelnr is 1-based
        if 0 < panelnr <= len(application.panels):
            application.activate_panel(application.panels[panelnr-1])


@registered_command
class EnterCommandMode(Command):
    """Start the command input field"""
    NAME = 'enter-command'

    def execute(self, context, text=""):
        if context.application.command_input is not None:
            return
        context.application.make_command_input(text)
        context.application.previous_focus = context.panel
        context.application.activate_panel(context.application.command_input)
        context.application.paint()
        logger.debug(f"Enter command input")


@registered_command
class CancelCommandInput(Command):
    """Unfocus the command input field"""
    NAME = 'cancel-command'

    def execute(self, context):
        context.application.command_input.destroy()
        context.application.command_input = None
        context.application.activate_panel(context.application.previous_focus)
        context.application.previous_focus = None
        context.application.paint()
        logger.debug(f"cancel command input")


@registered_command
class Repaint(Command):
    """Enforce repainting of screen"""
    NAME = 'repaint'

    def execute(self, context):
        context.application.resize_panels()
        context.application.paint(True)
        context.application.screen.noutrefresh()


@registered_command
class EditMetadataExternally(Command):
    """Launch an external text editor to edit the metadata of the selected file"""
    NAME = 'edit-metadata-external'
    ACCEPT_IN = (DocPanel, FilePanel, EditorPanel)

    def execute(self, context):
        target = context.panel
        file = target.selected_path

        # ensure there is something to edit
        if file is None:
            logger.debug(f"No file to edit")
            context.application.error("Nothing selected")
            return

        sidecar, is_collection, store = context.application.get_editable_sidecar_file(file)
        if sidecar is None:
            return

        if sidecar == file:
            context.application.error(f"Cannot edit metadata of a file that's probably a metadata file")
            return

        logger.debug(f"Editing sidecar file {sidecar} (collection? {is_collection})")

        if sidecar.is_file():
            if is_collection:
                meta = store.get_for_collection(sidecar, '')
                meta = utils.collection_meta_as_writable(meta, sidecar.parent)
            else:
                meta = store.get(sidecar, '')
                meta.popall(metaindex.shared.IS_RECURSIVE, [])

        else:
            # get the metadata for the selected item as a dict
            results = context.application.cache.get(file)
            logger.debug(f"Cached metadata for {file}: {results}")
            meta = multidict.MultiDict()

            if len(results) > 0:
                for key, value in results[0][1].items():
                    if not key.startswith('extra.'):
                        continue
                    meta.add(key.split('.', 1)[1], value)

        # resolve the external editor
        editor = context.application.get_text_editor(True)
        if editor is None:
            return

        # create the temporary file to edit
        with tempfile.NamedTemporaryFile("w+t", encoding="utf-8", suffix=sidecar.suffix) as fh:
            store.store(meta, fh)
            fh.flush()
            original = fh.read()
            with ShellContext(context.application.screen):
                subprocess.run(editor + [fh.name], check=False)
            context.application.paint(True)
            fh.flush()
            fh.seek(0)
            new_content = fh.read()
            fh.seek(0)
            changed = new_content != original
            if is_collection:
                updated = store.get_for_collection(fh, '', file.parent)
                updated = utils.collection_meta_as_writable(updated, file.parent)
            else:
                updated = store.get(fh, '')
                updated.popall(metaindex.shared.IS_RECURSIVE, [])

        if changed:
            store.store(updated, sidecar)
            context.application.cache.refresh(file)

            if isinstance(target, DocPanel):
                target.search(target.query)
                target.jump_to(file)
            elif isinstance(target, EditorPanel):
                target.reload()


@registered_command
class SelectFileAndExit(Command):
    """Write the selected file path to the requested location and exit"""
    NAME = 'select-and-exit'
    ACCEPT_IN = (DocPanel, FilePanel)

    def execute(self, context, *args):
        target = context.panel
        if target.is_busy:
            return

        if not context.application.select_file_mode:
            return

        item = target.items[target.cursor]
        path = None
        if isinstance(target, DocPanel):
            path = pathlib.Path(item[-1][0])
        elif isinstance(target, FilePanel):
            path = item
        else:
            return

        logger.debug(f"Selected file {path}")
        context.application.selected_file = path

        context.application.execute_command('quit')


@registered_command
class CopyPathToClipboard(Command):
    """Copy the selected item to clipboard"""
    NAME = 'copy'
    ACCEPT_IN = (DocPanel, FilePanel)

    def execute(self, context, clipboard=None):
        source = context.panel
        if source.is_busy:
            return

        context.application.clear_clipboard(clipboard)
        for path in source.selected_paths:
            context.application.append_to_clipboard(path, clipboard)


@registered_command
class AppendPathToClipboard(Command):
    """Append the selected item path to clipboard"""
    NAME = 'append'
    ACCEPT_IN = (DocPanel, FilePanel)

    def execute(self, context, clipboard=None):
        source = context.panel
        if source.is_busy:
            return

        for path in source.selected_paths:
            context.application.append_to_clipboard(path, clipboard)


@registered_command
class ClearClipboard(Command):
    """Clear the clipboard of items to copy"""
    NAME = 'clear-clipboard'

    def execute(self, context, clipboard=None):
        context.application.clear_clipboard(clipboard)
        context.application.info(f"Clipboard {clipboard} cleared")


@registered_command
class RefreshPanel(Command):
    """Refresh the current panel"""
    NAME = 'refresh'
    ACCEPT_IN = (DocPanel, FilePanel)

    def execute(self, context):
        if context.panel.is_busy:
            return

        target = context.panel
        logger.debug(f"Refresh {target}")

        if isinstance(target, DocPanel):
            target.search(target.query)
        elif isinstance(target, FilePanel):
            target.change_path(target.path)
        target.paint(True)


@registered_command
class DeleteItem(Command):
    """Delete selected item"""
    ACCEPT_IN = (FilePanel,)
    NAME = 'rm'

    def execute(self, context):
        if context.panel.is_busy:
            return

        context.panel.delete(context.panel.selected_paths)


class BookmarkNameReader(InputLine):
    def __init__(self, application, prefix):
        height, width = application.size()
        super().__init__(application, width, (height-1, 0), prefix=prefix)

    def handle_key(self, key):
        self.app.activate_panel(self.app.previous_focus)
        self.destroy()
        if len(str(key)) == 1 and str(key) in string.ascii_letters:
            self.bookmark_selected(str(key))
        else:
            self.text = None
        self.app.paint(True)

    def bookmark_selected(self, mark):
        raise NotImplementedError("Must be implemented in subclasses")


class LoadBookmarkNameReader(BookmarkNameReader):
    def __init__(self, application):
        super().__init__(application, 'Bookmark to load: ')

    def bookmark_selected(self, mark):
        self.app.hide_key_help()
        self.app.load_bookmark(mark)


class CreateBookmarkNameReader(BookmarkNameReader):
    def __init__(self, application, panel):
        super().__init__(application, 'Letter for the bookmark: ')
        self.target_panel = panel

    def bookmark_selected(self, mark):
        BookmarkItem.create_bookmark(mark, self.app, self.target_panel)


@registered_command
class BookmarkItem(Command):
    """Bookmark the selected item"""
    NAME = 'mark'
    ACCEPT_IN = (FilePanel, DocPanel)

    def execute(self, context, mark=None):
        if context.panel.is_busy:
            return

        if mark is None:
            reader = CreateBookmarkNameReader(context.application, context.panel)
            context.application.activate_panel(reader)
            context.application.paint()
            return

        self.create_bookmark(mark, context.application, context.panel)

    @staticmethod
    def create_bookmark(mark, application, panel):
        item = panel.selected_path

        if isinstance(panel, DocPanel):
            path = panel.query
        elif isinstance(panel, FilePanel):
            path = panel.path
        else:
            logger.error(f"Programming error: BookmarkItem command does not accept {type(target)}")
            return

        application.save_bookmark(mark, panel, path, item)


@registered_command
class JumpToBookmark(Command):
    """Open and jump to bookmark"""
    NAME = 'jump-to-mark'

    def execute(self, context, mark=None):
        if len(context.application.bookmarks) == 0:
            context.application.error("No bookmarks defined")
            return

        if mark is None:
            lines = [(mark, str(loc[1])) for mark, loc in context.application.bookmarks.items()]
            context.application.key_help_panel = KeyHelpPanel(lines, context.application)
            context.application.key_help_panel.autosize()
            reader = LoadBookmarkNameReader(context.application)
            context.application.activate_panel(reader)
            context.application.paint()
            return

        context.application.load_bookmark(mark)


@registered_command
class ChangeLayout(Command):
    """Change the layout"""
    NAME = 'layout'

    def execute(self, context, layout=None):
        accepted = layouts.layouts()
        cls = layouts.get_layout(layout)
        if cls is None or layout not in accepted:
            context.application.error(f"Possible layouts are: {', '.join(accepted)}")
            return

        context.application.layout = cls(context.application)
        context.application.resize_panels()
        context.application.paint(True)

    def completion_options(self, context, *args):
        text = "" if len(args) == 0 else args[0]
        return [layout for layout in sorted(layouts.layouts()) if layout.startswith(text)]


@registered_command
class GoToLocation(Command):
    """Open the folder of this document"""
    NAME = 'go-to-location'
    ACCEPT_IN = (DocPanel, DetailPanel, EditorPanel)

    def execute(self, context):
        if context.panel.is_busy:
            return

        path = context.panel.selected_path
        logger.debug(f"Opening {path} in new file panel")

        context.application.execute_command('new-file-panel')

        filepanel = context.application.panels[-1]
        filepanel.jump_to(path)

        context.application.paint(True)


@registered_command
class LoadConfig(Command):
    """Load the given configuration file"""
    NAME = 'source'

    def execute(self, context, path=None):
        if path is None:
            context.application.error(f"No path to configuration file given")
            return

        path = pathlib.Path(path).expanduser()

        if not path.is_file():
            msg = f"{path} does not exist or is not a file."
            logger.error(msg)
            context.application.error(msg)
            return

        context.application.load_config_file(path, context)


@simple_command('bind')
def bind_command(context, *args):
    """Set the keybinding to this command"""
    if len(args) < 3:
        context.application.error("Usage: bind scope keys command [help text]")
        return

    scope = args[0]
    keys = parse_key_sequence(args[1])
    context.application.keys.append((scope, keys, args[2:]))


@registered_command
class SetCommand(Command):
    """Set a configuration option"""
    NAME = 'set'

    def completion_options(self, context, *args):
        text = args[0] if len(args) > 0 else ""
        scope = ""
        name = ""

        options = []
        for group in sorted(context.application.configuration.conf.sections()):
            for option in sorted(context.application.configuration.conf[group]):
                full = group + "." + option
                if full.startswith(text):
                    options.append(group + "." + option)

        return options

    def execute(self, context, name=None, *values):
        if name is None:
            context.application.error("Usage: set name [value]")
            return

        if '.' in name:
            scope, name = name.split('.', 1)
        else:
            scope = 'all'

        if len(values) == 0:
            value = context.application.configuration.get(scope, name, '')
            context.application.info(f"{name}: {value}")
            return

        value = ' '.join(values)

        context.application.configuration.set(scope, name, value)
        context.application.configuration_changed((scope, name))


@simple_command('select', (FilePanel,))
def select_command(context):
    """Toggle whether or not the current item is selected"""
    if context.panel.is_busy:
        return

    item = context.panel.selected_item
    if item in context.panel.multi_selection:
        context.panel.multi_selection.remove(item)
    else:
        context.panel.multi_selection.append(item)
    context.panel.handle_key(context.panel.SCROLL_NEXT[0])


@registered_command
class ClearSelection(Command):
    """Unselect all selected items"""
    NAME = 'clear-selection'
    ACCEPT_IN = (FilePanel, DocPanel)

    def execute(self, context):
        if context.panel.is_busy:
            return

        indexes = [idx for idx, item in enumerate(context.panel.items) if item in context.panel.multi_selection]
        context.panel.multi_selection = []
        for index in indexes:
            context.panel.paint_item(index)


@registered_command
class InvertSelection(Command):
    """Invert selection"""
    NAME = 'invert-selection'
    ACCEPT_IN = (FilePanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return

        context.panel.multi_selection = [item for item in context.panel.items if item not in context.panel.multi_selection]
        context.panel.paint()


@registered_command
class RunOCR(Command):
    """Run OCR for this document"""
    NAME = 'ocr'
    ACCEPT_IN = (DocPanel, FilePanel, EditorPanel)

    def execute(self, context, languages=None):
        if context.panel.is_busy:
            return

        kwargs = {}
        if languages is not None:
            kwargs = {'languages': languages}

        path = context.panel.selected_path

        context.panel.run_blocking(self.run_ocr, context, path, **kwargs)
        context.application.paint(True)

    def run_ocr(self, blocker, context, path, **kwargs):
        blocker.title(f"Running OCR on {path.name}")
        success, fulltext = utils.do_ocr(path, **kwargs)

        if not success:
            logger.debug(f"OCR indexer returned nothing")
            context.application.info("OCR did not find anything useful")
            return

        meta = context.application.cache.get(path, False)
        if len(meta) == 0:
            meta = multidict.MultiDict()
        else:
            meta = meta[0][1]

        meta.add('ocr.fulltext', fulltext)
        logger.debug(f"New metadata to write: {meta}")
        context.application.cache.insert(path, meta)

        if isinstance(context.panel, (DocPanel,)):
            context.application.callbacks.put((context.panel,
                                               lambda: context.panel.search(context.panel.query)))
        elif isinstance(context.panel, (EditorPanel,)):
            context.panel.reload()
            context.application.callbacks.put((context.panel,
                                               lambda: context.panel.paint(True)))


@registered_command
class RunIndexers(Command):
    """Run indexers on the selected document"""
    NAME = 'index'
    ACCEPT_IN = (DocPanel, FilePanel, EditorPanel)

    def execute(self, context):
        if context.panel.is_busy:
            return

        path = context.panel.selected_path
        context.panel.run_blocking(self.run_indexers, context, path)
        context.application.paint(True)

    def run_indexers(self, blocker, context, path):
        blocker.title(f"Running indexers on {path.name}")
        
        item = context.panel.selected_item

        if path.is_dir():
            paths = context.application.cache.cache.find_indexable_files([path])
        else:
            paths = [path]

        results = metaindex.indexer.index_files(paths, 1, metaindex.ocr.Dummy(), False)
        if len(results) == 0:
            return

        blocker.progress((0, len(results)))

        # merge the results
        for idx, result in enumerate(results):
            lpath, success, extra = results[0]

            base = context.application.cache.get(lpath, False)
            if len(base) == 0:
                info = multidict.MultiDict()
            else:
                info = base[0][1]

            if not success:
                logger.debug(f"Indexer did not succeed on {lpath}")
                continue
            logger.debug(f"Indexer found something for {lpath}")
            
            # extend the cached metadata with the newly indexed data
            newly_added = False
            for key in set(extra.keys()):
                for value in extra.getall(key):
                    if value in info.getall(key, []):
                        continue
                    info.add(key, value)
                    newly_added = True

            if not newly_added:
                logger.debug("Nothing new here")
                return

            context.application.cache.insert(lpath, info)
            blocker.progress((idx+1, len(results)))

        if isinstance(context.panel, (DocPanel,)):
            context.panel.search(context.panel.query)
            context.application.callbacks.put((context.panel,
                                               lambda: context.panel.jump_to(item)))
        elif isinstance(context.panel, (EditorPanel,)):
            context.application.callbacks.put((context.panel,
                                               lambda: context.panel.reload()))




import curses
import collections
import tempfile
import subprocess

import multidict

from metaindex import shared
import metaindex.cache
import metaindex.indexer

from cursedspace import Key, InputLine, ShellContext

from metaindexmanager import command
from metaindexmanager import utils
from metaindexmanager.utils import logger
from metaindexmanager.panel import ListPanel
from metaindexmanager.detailpanel import DetailPanel
from metaindexmanager.docpanel import DocPanel
from metaindexmanager.filepanel import FilePanel


Change = collections.namedtuple("Change", ['index', 'new_value', 'prefix', 'tag', 'old_value'])
Insert = collections.namedtuple("Insert", ['prefix', 'tag', 'value'])
Delete = collections.namedtuple("Delete", ['index', 'prefix', 'tag', 'value'])
GroupedChange = collections.namedtuple("GroupedChange", ['changes'])

Line = collections.namedtuple("Line", ['group', 'prefix', 'tag', 'value'])
Header = collections.namedtuple("Header", ['group', 'title', 'prefix', 'tag', 'value'], defaults=['', '', ''])


class EditorLine(InputLine):
    def __init__(self, panel, *args, text=None, **kwargs):
        y = panel.pos[0] + panel.cursor - panel.offset + 1
        x = panel.pos[1] + panel.columns[0] + 1
        if text is None:
            text = panel.selected_line.value or ''
        super().__init__(panel.app, panel.columns[1], (y, x), text=text, background='░') # background='░.')

        self.item = panel.selected_line
        self.parent = panel
        self.app.previous_focus = panel
        self.app.current_panel = self
        self.original_text = self.text

        logger.debug(f"Enter tag edit mode for {self.item} (text: '{text}')")

    def handle_key(self, key):
        if key in [Key.ESCAPE, "^C"]:
            self.destroy()

        elif key in [Key.RETURN, "^S"]:
            self.destroy()
            if self.text != self.original_text:
                idx = self.parent.items.index(self.item)
                self.parent.changed(Change(idx, self.text,
                                           self.item.prefix,
                                           self.item.tag,
                                           self.item.value))

        else:
            super().handle_key(key)

    def destroy(self):
        super().destroy()
        self.parent.editor = None
        self.app.current_panel = self.parent
        self.parent.paint_item(self.parent.cursor)


class EditorPanel(ListPanel):
    SCOPE = 'editor'
    SPACING = 3

    CONFIG_ICON_MULTILINE = 'multiline-indicator'
    CONFIG_ICON_CUTOFF = 'cutoff-indicator'

    def __init__(self, filepath, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.item = filepath
        self.columns = []
        # the history of changes
        self.changes = []
        # the pointer into the history of changes, usually points beyond the end of the list of changes
        self.change_ptr = 0
        # list of items directly after reload
        self.unchanged_items = []
        self.editor = None
        self.metadata = multidict.MultiDict()

        self._multiline_icon = None
        self._cutoff_icon = None
        self.configuration_changed()
        
        self.reload()
        self.cursor = 1

    @property
    def selected_path(self):
        return self.item

    @property
    def selected_paths(self):
        return [self.item]

    @property
    def selected_item(self):
        return str(self.item)

    @property
    def selected_line(self):
        if 0 <= self.cursor < len(self.items):
            return self.items[self.cursor]
        return None

    def open_selected(self):
        if self.item is None:
            return
        self.app.open_file(self.item)

    def changed(self, change):
        logger.debug(f"Added change to stack: {change}")
        self.changes = self.changes[:self.change_ptr]
        self.changes.append(change)
        self.change_ptr = len(self.changes)
        self.rebuild_items()
        self.rebuild_columns()
        self.scroll()
        self.paint(True)

    def paint(self, clear=False):
        super().paint(clear)

        if self.change_ptr > 0:
            self.win.addstr(0, 1, " Modified ")
            self.win.noutrefresh()

    def title(self):
        return str(self.item)

    def focus(self):
        y, x = super().focus()
        if self.editor is not None:
            self.editor.focus()
        else:
            self.win.move(y, x+self.columns[0])
            self.win.noutrefresh()

    def multiline_edit(self):
        logger.debug(f"start multi-line editing {self.selected_line}")
        item = self.selected_line
        if not isinstance(item, Line):
            return
        original = item.value
        new_content = original
        can_edit = item.prefix == 'extra'

        editor = self.app.get_text_editor(True)
        if editor is None:
            return

        with tempfile.NamedTemporaryFile("w+t", encoding="utf-8", suffix='.txt') as fh:
            fh.write(original)
            fh.flush()
            with ShellContext(self.app.screen):
                subprocess.run(editor + [fh.name])
            self.app.paint(True)

            # only actually apply changes when editing the 'extra' tags
            if can_edit:
                fh.flush()
                fh.seek(0)
                new_content = fh.read()
        logger.debug(f"Can change? {can_edit} -- has changes? {new_content != original}")

        if can_edit and new_content != original:
            self.changed(Change(self.cursor, new_content, item.prefix, item.tag, original))

    def start_edit(self, text=None):
        logger.debug(f"start editing {self.selected_line}")
        if self.editor is not None:
            self.editor.destroy()
            del self.editor

        if (text is not None and '\n' in text) or (text is None and '\n' in self.selected_line.value):
            self.multiline_edit()
        else:
            self.editor = EditorLine(self, text=text)
            self.app.paint(True)

    def cancel_edit(self):
        if self.editor is None:
            return
        self.editor.destroy()
        del self.editor
        self.editor = None
        self.paint(True)

    def resize(self, *args):
        super().resize(*args)
        self.rebuild_columns()

    def reset(self):
        self.changes = []
        self.change_ptr = 0
        self.reload()
        self.paint(True)

    def save(self, blocker):
        logger.debug(f"The file is {self.selected_path.name}")
        blocker.title(f"Saving changes to {self.selected_path.name}")

        # Read in the current sidecar file, if it exists
        collection_extra = None
        sidecar_file, is_collection, store = self.app.get_editable_sidecar_file(self.selected_path)

        if sidecar_file is None:
            self.app.error(f"No usable metadata storage available")
            return

        if sidecar_file.exists():
            if is_collection:
                collection_extra = store.get_for_collection(sidecar_file, prefix='')

                collection_extra = utils.collection_meta_as_writable(collection_extra, sidecar_file.parent)
                logger.debug(f"Read collection metadata: {collection_extra}")

                if self.selected_path.name in collection_extra:
                    extra = collection_extra.pop(self.selected_path.name)
                else:
                    extra = multidict.MultiDict()
                logger.debug(f"Extra for {self.selected_path.name} is: {extra}")

            else:
                extra = store.get(sidecar_file, prefix='')
                # remove control data
                extra.popall(shared.IS_RECURSIVE, [])
        else:
            if is_collection:
                collection_extra = multidict.MultiDict()
            extra = multidict.MultiDict()

        # apply all changes to the extra metadata
        for change in self.expand_changes():
            logger.debug(f" ... processing {change}")
            prefix = ''
            if change.prefix != 'extra':
                prefix = change.prefix + '.'

            if isinstance(change, Change):
                values = extra.popall(prefix + change.tag, None)
                if values is None:
                    values = []

                applied = False
                for value in values:
                    if value == change.old_value and not applied:
                        extra.add(prefix + change.tag, change.new_value)
                        applied = True
                    else:
                        extra.add(prefix + change.tag, value)

                if not applied:
                    logger.info(f"Change to {prefix + change.tag} is actually an insert because of different sources")
                    extra.add(prefix + change.tag, change.new_value)

            elif isinstance(change, Insert):
                extra.add(prefix + change.tag, change.value)
            
            elif isinstance(change, Delete):
                values = extra.popall(prefix + change.tag, None)
                if values is None:
                    logger.warning(f"Skipping deletion of {prefix + change.tag}: not found")
                    continue
                if change.value in values:
                    values.remove(change.value)
                for value in values:
                    extra.add(prefix + change.tag, value)

        # 'store' will not remove the 'extra.' prefix per tag,
        # so we have to do that here
        for key in set(extra.keys()):
            if not key.startswith('extra.'):
                continue
            values = extra.popall(key)
            _, key = key.split('.', 1)
            for value in values:
                extra.add(key, value)
        logger.debug(f"Writing new metadata: {extra}")

        # save the extra metadata to the sidecar file
        if is_collection:
            collection_extra[self.selected_path.name] = extra
            logger.debug(f"Updating collection sidecar to {collection_extra}")
            store.store(collection_extra, sidecar_file)
        else:
            store.store(extra, sidecar_file)

        # reload the cache
        with ShellContext(self.app.screen):
            self.app.cache.refresh(self.selected_path)
        self.app.paint(True)

        # reset
        self.reset()

    def reload(self):
        filepath = str(self.item)
        metadata = [entry.metadata for entry in self.app.cache.get(self.item) if entry.path == self.item]

        if len(metadata) == 0:
            self.metadata = multidict.MultiDict({'filename': str(self.selected_path.name)})
        else:
            self.metadata = metadata[0]

        self.rebuild_items()
        self.rebuild_columns()

    def rebuild_columns(self):
        if len(self.items) > 1:
            self.columns = [max([1 if isinstance(row, Header) else len(str(row.tag))+self.SPACING
                                 for row in self.items]), 0]
            self.columns[1] = self.dim[1] - self.columns[0] - 2
        else:
            half_w = self.dim[1]//2
            self.columns = [half_w, self.dim[1] - half_w]

    def rebuild_items(self):
        self.items = []
        self.unchanged_items = []

        if len(self.metadata) == 0:
            self.unchanged_items = []
            self.cursor = 0
            return

        keys = list(set(self.metadata.keys()))
        keys.sort(key=lambda k: [not k.startswith('extra.'), '.' in k, k.lower()])

        # prepare the unchanged items
        for key in keys:
            displaykey = key
            prefix = 'general'
            if '.' in key:
                prefix, displaykey = key.split('.', 1)
            for value in self.metadata.getall(key, []):
                self.unchanged_items.append(Line(prefix, prefix, displaykey, value))

        # apply all changes in order to the point where we are
        self.items = self.unchanged_items[:]
        for change in [None] + self.expand_changes():
            if isinstance(change, Change):
                original = self.items[change.index]
                self.items[change.index] = Line(original.group,
                                                original.prefix,
                                                original.tag,
                                                change.new_value)
            elif isinstance(change, Insert):
                group = change.prefix
                if len(group) == 0:
                    group = 'general'
                self.items.append(Line(group, change.prefix, change.tag, change.value))
            elif isinstance(change, Delete):
                self.items = self.items[:change.index] + self.items[change.index+1:]

            self.items = [i for i in self.items if isinstance(i, Line)]
            self.items += [Header(g, g.title())
                           for g in set([i.group for i in self.items])]

            self.items.sort(key=lambda k: [k.group != 'extra', k.group, isinstance(k, Line), k.tag.lower(), self.app.humanize(k.value).lower()])

    def do_paint_item(self, y, x, maxwidth, is_selected, item):
        if isinstance(item, Header):
            self.win.addstr(y, x, item.title[:self.dim[1]-x-2], curses.A_BOLD)
        else:
            for colidx, text in enumerate([item.tag, item.value]):
                self.win.addstr(y, x, " "*self.columns[colidx])
                maxlen = self.dim[1]-x-2

                if text is None:
                    text = ''
                if colidx == 1 and is_selected and self.editor is not None:
                    self.editor.paint()
                else:
                    # make it a human-readable string
                    text = self.app.humanize(text)

                    # multi-line is special
                    is_multiline = '\r' in text or '\n' in text
                    text = utils.first_line(text)
                    if is_multiline:
                        text += ' ' + self._multiline_icon

                    # shorten the text to visible width
                    shortened = text[:maxlen]

                    if len(shortened) < len(text):
                        icon = self._cutoff_icon
                        if is_multiline:
                            icon = self._multiline_icon
                        shortened = shortened[:-1-len(icon)] + ' ' + icon

                    self.win.addstr(y, x, shortened[:maxlen])
                x += self.columns[colidx]

    def configuration_changed(self, name=None):
        super().configuration_changed(name)

        changed = False

        if name is None or name == self.CONFIG_ICON_MULTILINE:
            new_value = self.app.configuration.get(self.SCOPE, self.CONFIG_ICON_MULTILINE, '…')
            changed = self._multiline_icon != new_value
            self._multiline_icon = new_value

        if name is None or name == self.CONFIG_ICON_CUTOFF:
            new_value = self.app.configuration.get(self.SCOPE, self.CONFIG_ICON_CUTOFF, '→')
            changed = self._multiline_icon != new_value
            self._cutoff_icon = new_value

        if changed:
            if self.win is not None:
                self.scroll()
                self.paint(True)

    def add_tag(self, name):
        if self.editor is not None:
            self.cancel_edit()
        tagname = f'extra.{name}'

        self.changed(Insert('extra', name, ''))

        for nr, item in enumerate(self.items):
            if not isinstance(item, Line):
                continue
            if item.prefix == 'extra' and item.tag == name and item.value == '':
                self.cursor = nr
        self.scroll()

        self.paint(True)

    def remove_tag(self, line):
        if line not in self.items or isinstance(line, str):
            return

        if self.editor is not None:
            self.cancel_edit()

        idx = self.items.index(line)
        self.changed(Delete(idx, line.prefix, line.tag, line.value))

    def undo(self):
        if self.change_ptr <= 0:
            return

        self.change_ptr -= 1
        self.rebuild_items()
        self.rebuild_columns()
        self.scroll()
        self.paint(True)

    def redo(self):
        if self.change_ptr >= len(self.changes):
            return
        self.change_ptr += 1
        self.rebuild_items()
        self.rebuild_columns()
        self.scroll()
        self.paint(True)

    def expand_changes(self):
        if len(self.changes[:self.change_ptr]) == 0:
            return []
        return sum([change.changes if isinstance(change, GroupedChange) else [change]
                    for change in self.changes[:self.change_ptr]], start=[])


@command.registered_command
class EditMetadata(command.Command):
    """Edit metadata of the selected file"""
    NAME = 'edit-metadata'
    ACCEPT_IN = (DocPanel, FilePanel, DetailPanel)

    def execute(self, context):
        if context.panel.is_busy:
            return
        item = context.panel.selected_path

        if not item.is_file():
            context.application.error(f"{item.stem} is not a file")
            return

        panel = EditorPanel(item, context.application)
        context.application.add_panel(panel)
        context.application.activate_panel(panel)


@command.registered_command
class EnterEditMode(command.Command):
    """Start editing the metadata"""
    NAME = 'edit-mode'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        item = context.panel.selected_line

        if not isinstance(item, Line):
            return

        if item.prefix == 'extra' or (isinstance(item.value, str) and ('\n' in item.value or '\r' in item.value)):
            context.panel.start_edit()


@command.simple_command("edit-multiline", (EditorPanel,))
def edit_multiline_command(context):
    target = context.panel
    if target.is_busy:
        return

    context.panel.multiline_edit()


@command.simple_command('add-tag', (EditorPanel,))
def execute(context, name=None):
    """Add a new metadata field"""
    if context.panel.is_busy:
        return
    if name is None:
        context.application.error(f"Usage: add-attr name")
        return

    context.panel.add_tag(name)


@command.registered_command
class AddValueForAttribute(command.Command):
    """Add a new metadata value for this field"""
    NAME = 'add-value'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        item = context.panel.selected_line

        if not isinstance(item, Line) or item.prefix != 'extra':
            return

        context.panel.add_tag(item.tag)


@command.registered_command
class ReplaceValueForAttribute(command.Command):
    """Replace the selected metadata value"""
    NAME = 'replace-value'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        item = context.panel.selected_line

        if not isinstance(item, Line) or item.prefix != 'extra':
            return

        context.panel.start_edit(text='')

@command.registered_command
class RemoveAttribute(command.Command):
    """Remove the selected metadata field"""
    NAME = 'del-tag'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        item = context.panel.selected_line

        if isinstance(item, Header):
            context.application.error(f"Selected field cannot be deleted")
            return
        if item.prefix != 'extra':
            # TODO support the null override of values
            context.application.error(f"Selected field cannot be deleted")
            return
        context.panel.remove_tag(context.panel.selected_line)


@command.registered_command
class ResetEdits(command.Command):
    """Reset all unsaved changes"""
    NAME = 'reset'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        context.panel.reset()


@command.registered_command
class SaveChanges(command.Command):
    """Save metadata changes"""
    NAME = 'write'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        context.panel.run_blocking(context.panel.save)
        context.application.paint(True)


@command.registered_command
class UndoChange(command.Command):
    """Undo the previous change"""
    NAME = 'undo-change'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        context.panel.undo()


@command.registered_command
class RedoChange(command.Command):
    """Redo the next change (i.e. undo the undo)"""
    NAME = 'redo-change'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        context.panel.redo()


@command.registered_command
class UndoAllChanges(command.Command):
    """Undo all changes"""
    NAME = 'undo-all-changes'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return
        if context.panel.change_ptr <= 0:
            return
        context.panel.change_ptr = 1
        context.panel.undo()


@command.simple_command("copy-tag", (EditorPanel,))
def copy_tag_command(context, clipboard=None):
    """Copy the selected tag and value to clipboard"""
    source = context.panel
    if source.is_busy:
        return

    context.application.clear_clipboard(clipboard)

    line = source.selected_line

    if line is None or isinstance(line, Header):
        return

    context.application.append_to_clipboard((line,), clipboard)


@command.simple_command("copy-append-tag", (EditorPanel,))
def copy_append_tag_command(context, clipboard=None):
    """Add the selected tag and value to clipboard"""
    source = context.panel
    if source.is_busy:
        return

    line = source.selected_line

    if line is None or isinstance(line, Header):
        return

    context.application.append_to_clipboard((line,), clipboard)


@command.simple_command("paste-tag", (EditorPanel,))
def paste_tag_command(context, clipboard=None):
    target = context.panel
    if target.is_busy:
        return

    items = [item for item in context.application.get_clipboard_content(clipboard)
             if isinstance(item, (Line,))]
    if len(items) == 0:
        return

    changes = [Insert('extra', line.tag, line.value) for line in items]
    grouped = GroupedChange(changes)
    target.changed(grouped)


@command.registered_command
class RunRules(command.Command):
    """Run tag rules on this document"""
    NAME = 'rules'
    ACCEPT_IN = (EditorPanel,)

    def execute(self, context):
        if context.panel.is_busy:
            return

        if metaindex.indexer.get('rule-based') is None:
            context.application.error("Rule based indexer not found")
            return

        path = context.panel.selected_path
        context.panel.run_blocking(self.run_rules, context, path)
        context.application.paint(True)

    def run_rules(self, blocker, context, path):
        blocker.title(f"Running rules on {path.name}")
        
        item = context.panel.selected_item

        base = context.application.cache.get(path, False)
        if len(base) == 0:
            info = metaindex.cache.Cache.Entry(path)
        else:
            info = base[0]

        fulltext = shared.get_all_fulltext(info.metadata)
        if len(fulltext) == 0:
            # this will also run the rule-based indexer
            logger.debug(f"No fulltext available, running indexer on {path}")
            results = metaindex.indexer.index_files([path],
                                                    1,
                                                    metaindex.ocr.TesseractOCR(True),
                                                    True,
                                                    context.application.metaindexconf)
            if len(results) == 0:
                logger.debug(f"Indexers returned no results")
                return
            _, success, base = results[0]

            info = metaindex.cache.Cache.Entry(path, base, shared.get_last_modified(path))

        else:
            # there is some fulltext, just rerun the rules
            logger.debug(f"Fulltext is already here: {len(fulltext)}")

            cache = metaindex.indexer.IndexerCache(metaindex.ocr.Dummy(),
                                                   False,
                                                   context.application.metaindexconf,
                                                   {},
                                                   info)
            indexer = metaindex.indexer.get('rule-based')(cache)
            success, extra = indexer.run(path, info.metadata.copy(), info)

            if not success:
                logger.debug(f"Indexer did not succeed")
                return
            
            # extend the cached metadata with the newly indexed data
            new_info = False
            for key in set(extra.keys()):
                for value in extra.getall(key):
                    if value in info.metadata.getall(key, []):
                        continue
                    info.metadata.add(key, value)
                    new_info = True

            if not new_info:
                logger.debug("Nothing new here")
                return

        context.application.cache.insert(path, info.metadata)

        context.application.callbacks.put((context.panel,
                                           lambda: context.panel.reload()))




import datetime
import pathlib
import shlex
import curses

from cursedspace import Panel

from metaindexmanager import command
from metaindexmanager import utils
from metaindexmanager.utils import logger
from metaindexmanager.panel import ListPanel


DEFAULT_COLUMNS = "title, filename, tags+, mimetype"


class DocPanel(ListPanel):
    SCOPE = 'documents'
    
    CONFIG_COLUMNS = 'columns'

    def __init__(self, *args, searchterm='', **kwargs):
        super().__init__(*args, **kwargs)
        self.col_widths = []
        self.fieldkeys = []
        self.query = None
        self.post_load = None
        self.results = []

        self.configuration_changed()
        self.search(searchterm)

    @property
    def synonyms(self):
        return self.app.metaindexconf.synonyms

    def title(self):
        return str(self.query)

    def configuration_changed(self, name=None):
        super().configuration_changed(name)

        changed = False

        if name is None or name == self.CONFIG_COLUMNS:
            keys = self.app.configuration.list(self.SCOPE,
                                               self.CONFIG_COLUMNS,
                                               DEFAULT_COLUMNS,
                                               separator=' ')

            changed = self.fieldkeys != keys
            self.fieldkeys = keys

        if changed:
            self.rebuild_items()
            self.calculate_grid_size()
            if self.win is not None:
                self.scroll()
                self.paint(True)

    def search(self, query):
        logger.debug(f"DocPanel: search for {query}")
        self.query = query
        self.run_in_background(self.do_search)

    def do_search(self):
        if not self.app.cache.is_initialized:
            self.app.cache.wait_for_reload()
        self.results = list(self.app.cache.find(self.query))
        self.cursor = 0
        self.rebuild_items()
        self.calculate_grid_size()
        if self.post_load is not None:
            self.post_load()
            self.post_load = None
        self.app.callbacks.put((self, lambda: True))

    def rebuild_items(self):
        self.items = [self.make_row(result) for result in self.results]
        self.items = [item for item in self.items if sum([len(field) for field in item[:-1]]) > 0]
        self.items.sort(key=lambda i: [len(i[0]) == 0] + [v.lower() for v in i[:-1]])

    def make_row(self, result):
        fields = []
        for key in self.fieldkeys:
            multivalue = False
            if key.endswith('+'):
                multivalue = True
                key = key[:-1]

            if key == 'icon':
                values = [utils.get_ls_icon(pathlib.Path(result[0]), {})]
            else:
                values = []
                for expandedkey in set(self.synonyms.get(key, [key])):
                    values += result[1].getall(expandedkey, [])

            if len(values) == 0:
                fields.append("")
            elif len(values) > 1 and multivalue:
                fields.append(', '.join([self.app.humanize(v) for v in values if len(v) > 0]))
            else:
                fields.append(self.app.humanize(values[0]))
            fields[-1] = fields[-1].replace("\n", " ").replace("\r", " ").replace("\t", " ")

        fields.append(result)
        return fields

    def calculate_grid_size(self):
        if self.win is None:
            return
        # calculate the spread of the grid prior to calling super.paint
        maxheight, maxwidth = self.dim
        margin = 0
        if self.border & Panel.BORDER_LEFT != 0:
            margin += 1
        if self.border & Panel.BORDER_RIGHT != 0:
            margin += 1

        if len(self.fieldkeys) == 0:
            self.col_widths = [0]
            return

        # recalculate how wide each column is
        equal_colwidth = (maxwidth - margin)//len(self.fieldkeys)
        self.col_widths = [max([0] + [max(1 if fkey == 'icon' else equal_colwidth,
                                      len(row[fkidx]))+1 for row in self.items])
                           for fkidx, fkey in enumerate(self.fieldkeys)]

        # shrink each column if its wider than the equal_colwidths, start from
        # the last column (it's probably the least important)
        if len(self.col_widths) > 2:
            to_shrink = len(self.col_widths)-1
            while sum(self.col_widths) > maxwidth-margin and to_shrink >= 0:
                self.col_widths[to_shrink] = min(equal_colwidth, self.col_widths[to_shrink])
                to_shrink -= 1
        # expand/shrink the last column to the rest of the available space
        self.col_widths[-1] = maxwidth - margin - sum(self.col_widths[:-1])

    def resize(self, *args):
        super().resize(*args)
        self.calculate_grid_size()

    def paint(self, clear=False):
        if self.is_busy:
            Panel.paint(self, clear)
            x, y = 0, 0
            if (self.border & self.BORDER_TOP) != 0:
                y += 1
            if (self.border & self.BORDER_LEFT) != 0:
                x += 1
            self.win.addstr(y, x, "...")
            self.win.noutrefresh()
        else:
            super().paint(clear)

    def do_paint_item(self, y, x, maxwidth, is_selected, item):
        assert sum(self.col_widths) < self.dim[1]
        attr = curses.A_STANDOUT if is_selected else curses.A_NORMAL
        self.win.addstr(y, x, " "*maxwidth, attr)
        for column in range(len(self.fieldkeys)):
            if self.col_widths[column] < 2:
                continue
            text = item[column][:self.col_widths[column]]
            if column < len(self.fieldkeys) - 1 and len(text) >= self.col_widths[column]:
                text = text[:-1] + ' '
            assert len(text) + x < self.dim[1]
            try:
                self.win.addstr(y, x, text, attr)
            except:
                pass
            x += self.col_widths[column]

    def jump_to(self, item, path=None):
        if path is None:
            path = self.query

        if path != self.query:
            self.search(path)

        if self.is_busy:
            self.post_load = lambda: self.do_jump_to(item)
            return

        self.do_jump_to(item)

    def do_jump_to(self, item):
        targetitem = str(item)

        for cursor, item in enumerate(self.items):
            if item[-1][0] == targetitem:
                self.cursor = cursor
                break

        self.scroll()
        if not self.is_busy:
            self.paint()

    @property
    def selected_path(self):
        result = self.selected_item
        if result is not None:
            result = pathlib.Path(result[-1][0])
        return result

    @property
    def selected_paths(self):
        return [pathlib.Path(i[-1][0]) for i in self.selected_items]

    def open_selected(self):
        if self.selected_item is None:
            self.app.error("Nothing selected")
            return
        path = pathlib.Path(self.selected_item[-1][0])
        self.app.open_file(path)


@command.registered_command
class NewMetaPanel(command.Command):
    """Create a new metadata panel"""
    NAME = 'new-documents-panel'

    def execute(self, context):
        if context.panel.is_busy:
            return

        panel = DocPanel(context.application)
        context.application.add_panel(panel)
        context.application.activate_panel(panel)


@command.registered_command
class Search(command.Command):
    """Search documents with this metaindex search term"""
    NAME = 'search'
    ACCEPT_IN = (DocPanel,)

    def execute(self, context, *args):
        if context.panel.is_busy:
            return

        term = shlex.join(args)
        context.panel.search(term)
        context.panel.paint(clear=True)
        context.application.paint_focus_bar()


@command.registered_command
class SetColumns(command.Command):
    """Set the columns of this panel"""
    NAME = 'columns'
    ACCEPT_IN = (DocPanel,)

    def execute(self, context, *args):
        if len(args) == 0:
            context.application.info(f"columns = {' '.join(context.panel.fieldkeys)}")
            return

        if context.panel.is_busy:
            return

        context.panel.fieldkeys = args
        context.panel.rebuild_items()
        if context.panel.cursor >= len(context.panel.items):
            context.panel.cursor = len(context.panel.items)-1
        if context.panel.win is not None:
            context.panel.calculate_grid_size()
            context.panel.scroll()
            context.panel.paint(True)



import threading
import curses

from cursedspace import ScrollPanel, Key

from metaindexmanager.utils import logger


NEXT_ITEM = 'next-item'
PREV_ITEM = 'previous-item'
NEXT_PAGE = 'next-page'
PREV_PAGE = 'previous-page'
TO_START = 'go-to-start'
TO_END = 'go-to-end'
LIST_SCROLL_COMMANDS = [
    NEXT_ITEM,
    PREV_ITEM,
    NEXT_PAGE,
    PREV_PAGE,
    TO_START,
    TO_END,
]

OPTION_PAGING = 'paging'
OPTION_SCROLL_MARGIN = 'scroll-margin'


class BlockingTask(ScrollPanel):
    INDICATOR_1 = '-\\|/'
    # INDICATOR_2 = '.oOo'
    INDICATOR_2 = '▂▃▄▅▆▇▆▅▄▃▂'

    def __init__(self, app):
        super().__init__(app)
        self.border = ScrollPanel.BORDER_ALL
        self.target = None
        self.task = None
        self._title = "Blocking task"
        self._changed = False
        self._progress = None
        self._description = []
        self._status = None
        self.indicator = self.INDICATOR_1
        self.indicator_pos = (0, 0)
        self.indicator_idx = 0
        self.autoresize()

    def handle_key(self, key):
        return False

    def check(self):
        if self.task is None or not self.task.is_alive():
            return False

        if self._changed:
            self._changed = False
            self.paint(True)
        else:
            self.paint_indicator()
        
        threading.Timer(0.5, lambda: self.app.callbacks.put((self, self.check))).start()

        return True

    def paint(self, clear=False):
        super().paint(clear)

        height, width = self.dim

        text = self._title[:width-3]
        self.win.addstr(1, 1, text, curses.A_BOLD)

        if self._progress is not None:
            y = 3
            x = 2
            bar_w = width - 4
            if isinstance(self._progress, tuple):
                percent = float(self._progress[0])/self._progress[1]
            else:
                percent = self._progress/100.0
            percent = min(1.0, max(0.0, percent))
            completed_w = int(bar_w*percent)
            self._indicator_pos = (y, x + completed_w)
            self.win.addstr(y, x, "▒"*bar_w)
            self.win.addstr(y, x, "█"*completed_w)

        self.win.addstr(4, 1, " "*(width-2))
        if self._status is not None:
            self.win.addstr(4, 1, self._status[:width-2])

        self.win.noutrefresh()
        self.paint_indicator()

    def paint_indicator(self):
        self.win.addstr(self._indicator_pos[0], self._indicator_pos[1], self.indicator[self.indicator_idx])
        self.indicator_idx = (self.indicator_idx + 1) % len(self.indicator)
        self.win.noutrefresh()

    def autoresize(self):
        maxheight, maxwidth = self.app.size()
        width = min(maxwidth, max(20, 3*maxwidth//4))
        height = min(maxheight, max(5, maxheight//3))
        y = (maxheight-height)//2
        x = (maxwidth-width)//2
        logger.debug(f"Autoresizing blocking task window to {height}x{width} +{x}:{y}")
        if self._progress is None:
            self._indicator_pos = (0, width-3)
        self.resize(height, width)
        self.move(y, x)

    def status(self, status):
        """Update the status text"""
        self._status = status
        self._changed = True

    def title(self, title):
        """Set the title of the task"""
        self._title = title
        self._changed = True

    def progress(self, current):
        """Update the progress of the task

        current may be an int with the progress in percent (0..100)
        or it may be a tuple with (completed subtasks, number of subtasks)"""
        self._progress = current
        if self._progress is None:
            self.indicator = self.INDICATOR_1
        else:
            self.indicator = self.INDICATOR_2
        self._changed = True

    def destroy(self):
        super().destroy()
        self.app.blocking_task = None
        self.app.paint()
        return True

    def run(self):
        self.task = threading.Thread(target=self.do_run)
        self.task.start()
        # only paint if the task takes longer than 0.1
        self.task.join(0.1)
        logger.debug("Blocking task launched")

    def do_run(self):
        self.app.callbacks.put((self, self.check))
        try:
            self.target()
        except Exception as exc:
            logger.error(f"Blocking task failed: {exc}")
        self.app.callbacks.put((self, self.destroy))


class ListPanel(ScrollPanel):
    def __init__(self, application, *args, **kwargs):
        super().__init__(application, *args, **kwargs)
        self.reload_keybindings()
        self._is_busy = threading.Lock()
        self.multi_selection = []
    
    def reload_keybindings(self):
        self.SCROLL_NEXT = extract_key_sequence(self.app, NEXT_ITEM)
        self.SCROLL_PREVIOUS = extract_key_sequence(self.app, PREV_ITEM)
        self.SCROLL_NEXT_PAGE = extract_key_sequence(self.app, NEXT_PAGE)
        self.SCROLL_PREVIOUS_PAGE = extract_key_sequence(self.app, PREV_PAGE)
        self.SCROLL_TO_START = extract_key_sequence(self.app, TO_START)
        self.SCROLL_TO_END = extract_key_sequence(self.app, TO_END)

    def title(self):
        raise NotImplementedError()

    @property
    def selected_items(self):
        if len(self.multi_selection) == 0:
            return [self.selected_item]
        return self.multi_selection

    @property
    def is_busy(self):
        return self._is_busy.locked()

    def run_blocking(self, target, *args, **kwargs):
        """Run the function 'target' as a blocking task

        Pass *args, and **kwargs along to target.
        'target' must accept the blockingtask class as its first argument.
        """
        if self.app.blocking_task is not None:
            self.app.error("Another blocking task is already running")

        self.app.blocking_task = BlockingTask(self.app)
        self.app.blocking_task.target = lambda: target(self.app.blocking_task, *args, **kwargs)
        self.app.blocking_task.run()

    def run_in_background(self, target):
        """Run the function 'target' as a thread in the background
        """
        if self.is_busy:
            raise RuntimeError("Already busy")
        thread = threading.Thread(target=lambda: self._do_run_in_background(target))
        logger.debug(f"{self} starts a background task")
        thread.start()

    def _do_run_in_background(self, target):
        with self._is_busy:
            target()

    def handle_key(self, key):
        if self.is_busy:
            return

        return super().handle_key(key)

    def configuration_changed(self, name=None):
        changed = False

        if name == OPTION_PAGING or name is None:
            new_value = self.app.configuration.bool('all', OPTION_PAGING, str(self.SCROLL_PAGING))
            changed = new_value != self.SCROLL_PAGING
            self.SCROLL_PAGING = new_value

        if name == OPTION_SCROLL_MARGIN or name is None:
            new_value = self.app.configuration.number('all', OPTION_SCROLL_MARGIN, str(self.SCROLL_MARGIN))
            if new_value is not None and new_value >= 0:
                changed = changed or new_value != self.SCROLL_MARGIN
                self.SCROLL_MARGIN = new_value

        if changed and self.win is not None:
            self.scroll()
            self.paint(True)

    def focus(self):
        retval = super().focus()
        self.on_focus()
        return retval

    def on_close(self):
        pass

    def on_focus(self):
        pass


def extract_key_sequence(application, commandname):
    return [keys
            for scope, keys, cmd in application.keys
            if cmd[0] == commandname and scope == 'any']





import curses
import datetime
import sys
import os
import shlex
import logging
import shutil
import argparse
import subprocess
import pathlib
import traceback
import importlib
import queue
import configparser

import metaindex.configuration
import metaindex.logger
from metaindex import stores
from metaindex.cache import MemoryCache

import cursedspace
from cursedspace import colors
from cursedspace import Panel
from cursedspace import Key
from cursedspace import InputLine
from cursedspace import ShellContext

import metaindexmanager.commands as _
import metaindexmanager.command as _
from . import layouts
from . import utils
from .utils import logger
from .panel import LIST_SCROLL_COMMANDS
from .shared import ANY_SCOPE, ALL_SCOPE, PROGRAMNAME
from .command import resolve_command, Command
from .docpanel import DocPanel
from .filepanel import FilePanel
from .detailpanel import DetailPanel as _
from .commandinput import CommandInput, expand_part
from .keyhelppanel import KeyHelpPanel
from .editorpanel import EditorPanel


CONFIGPATH = metaindex.configuration.HOME / ".config" / PROGRAMNAME
DATAPATH = metaindex.configuration.HOME / ".local" / "share" / PROGRAMNAME

try:
    from xdg import BaseDirectory
    DATAPATH = pathlib.Path(BaseDirectory.save_data_path(PROGRAMNAME) or DATAPATH)
    CONFIGPATH = pathlib.Path(BaseDirectory.save_config_path(PROGRAMNAME) or CONFIGPATH)
except ImportError:
    BaseDirectory = None

ADDONSPATH = DATAPATH / "addons"
LOGFILE = DATAPATH / "ui.log"
HISTORYFILE = DATAPATH / "history.txt"
BOOKMARKFILE = DATAPATH / "bookmarks.txt"
CONFIGFILE = CONFIGPATH / "config.rc"

PREF_OPENER = 'opener'
DEFAULT_OPENER = 'xdg-open'
PREF_SIDECAR = 'preferred-sidecar-format'
DEFAULT_SIDECARS = '.json'
PREF_EXT_EDITOR = 'editor'


class Context:
    def __init__(self, **kwargs):
        self.panel = kwargs.get('panel', None)
        self.application = kwargs.get('application', None)


class YesNoConfirmation(InputLine):
    def __init__(self, callback, default, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.callback = callback
        self.default = default.lower()[:1]

    def handle_key(self, key):
        if key == Key.RETURN:
            text = self.text.strip().lower()
            if len(text) == 0:
                text = self.default
            if len(text) > 0 and text.startswith('y'):
                self.callback()
            try:
                self.win.addstr(0, 0, " "*self.dim[1])
            except:
                pass
            self.win.noutrefresh()
            self.destroy()

            self.app.current_panel = self.app.previous_focus
            self.app.paint()
        else:
            super().handle_key(key)


class Application(cursedspace.Application):
    DEFAULT_CLIPBOARD = ''

    def __init__(self, args):
        super().__init__()
        # key definition is quirky:
        # the mapping is 'sequence of keys' to 'command'
        # command can be either of these:
        # - 'command', name of a command (to be resolved and executed as is)
        # - '::command param1 param2', command with parameters to be executed as such
        # - ':command param1', command to be filled into the command line and the user has to complete it
        self.keys = [
            (ANY_SCOPE, ('q', ), ('close',)),
            (ANY_SCOPE, ('X', ), ('quit',)),
            (FilePanel.SCOPE, (Key.RIGHT,), ('open',)),
            (DocPanel.SCOPE, (Key.RIGHT,), ('open',)),
            (EditorPanel.SCOPE, (Key.RIGHT,), ('open',)),
            (FilePanel.SCOPE, (Key.LEFT,), ('go-to-parent',)),
            (ANY_SCOPE, (Key.TAB, ), ('next-panel',)),
            (ANY_SCOPE, ('g', 'n', 'f'), ('new-file-panel',)),
            (ANY_SCOPE, ('g', 'n', 'd'), ('new-documents-panel',)),
            (ANY_SCOPE, ('g', 'p'), ('::focus',)),
            (ANY_SCOPE, ('g', 't'), ('next-panel',)),
            (ANY_SCOPE, ('g', 'T'), ('previous-panel',)),
            (ANY_SCOPE, ('g', 'l'), ('go-to-location',)),
            (DocPanel.SCOPE, ('g', 's'), (':search',)),
            (DocPanel.SCOPE, (Key.F3, ), (':search',)),
            (DocPanel.SCOPE, ('/', ), (':search',)),
            (FilePanel.SCOPE, ('g', 'h'), ('::cd ~', "Go home")),
            (FilePanel.SCOPE, ('g', 'd'), ('details',)),
            (DocPanel.SCOPE, ('g', 'd'), ('details',)),
            (FilePanel.SCOPE, ('c', 'd'), (':cd',)),
            (FilePanel.SCOPE, ('d', 'D'), ('rm',)),
            (FilePanel.SCOPE, ('y', 'y'), ('copy',)),
            (FilePanel.SCOPE, ('y', 'a'), ('append',)),
            (FilePanel.SCOPE, ('!', ), (':shell',)),
            (DocPanel.SCOPE, ('y', 'y'), ('copy',)),
            (DocPanel.SCOPE, ('y', 'a'), ('append',)),
            (EditorPanel.SCOPE, ('y', 'y'), ('copy-tag',)),
            (EditorPanel.SCOPE, ('y', 'a'), ('copy-append-tag',)),
            (EditorPanel.SCOPE, ('p', 'p'), ('paste-tag',)),
            (EditorPanel.SCOPE, ('p', 'P'), ('paste-tag',)),
            (ANY_SCOPE, ('y', '0'), ('clear-clipboard',)),
            (FilePanel.SCOPE, ('z', 'h'), ('toggle-hidden',)),
            (ANY_SCOPE, ('z', 's'), ('::layout horizontal', 'Split layout')),
            (ANY_SCOPE, ('z', 't'), ('::layout tabbed', 'Tabbed layout')),
            (ANY_SCOPE, (':', ), ('enter-command',)),
            (ANY_SCOPE, ('^L', ), ('repaint',)),
            (DocPanel.SCOPE, ('e', ), ('edit-metadata',)),
            (DocPanel.SCOPE, ('E', ), ('edit-metadata-external',)),
            (FilePanel.SCOPE, ('e', ), ('edit-metadata',)),
            (FilePanel.SCOPE, ('E', ), ('edit-metadata-external',)),
            (EditorPanel.SCOPE, ('E', ), ('edit-metadata-external',)),
            (ANY_SCOPE, ('^R', ), ('refresh',)),
            (FilePanel.SCOPE, ('m', ), ('::mark',)),
            (DocPanel.SCOPE, ('m', ), ('::mark',)),
            (FilePanel.SCOPE, (Key.SPACE,), ('select',)),
            (FilePanel.SCOPE, ('u', 'v'), ('clear-selection',)),
            (FilePanel.SCOPE, ('v', ), ('invert-selection',)),
            (ANY_SCOPE, ("'", ), ('::jump-to-mark',)),
            (ANY_SCOPE, (Key.DOWN,), ('next-item',)),
            (ANY_SCOPE, (Key.UP,), ('previous-item',)),
            (ANY_SCOPE, (Key.PGUP,), ('previous-page',)),
            (ANY_SCOPE, (Key.PGDN,), ('next-page',)),
            (ANY_SCOPE, (Key.HOME,), ('go-to-start',)),
            (ANY_SCOPE, (Key.END,), ('go-to-end',)),
            (EditorPanel.SCOPE, (Key.RETURN,), ('edit-mode',)),
            (EditorPanel.SCOPE, ('i',), ('edit-mode',)),
            (EditorPanel.SCOPE, ('o',), ('add-value',)),
            (EditorPanel.SCOPE, ('c',), ('replace-value',)),
            (EditorPanel.SCOPE, ('a',), (':add-tag', 'Add a new tag')),
            (EditorPanel.SCOPE, ('d', 'd'), ('del-tag',)),
            (EditorPanel.SCOPE, ('u',), ('undo-change',)),
            (EditorPanel.SCOPE, ('U',), ('undo-all-changes',)),
            (EditorPanel.SCOPE, ('^R',), ('redo-change',)),
            (EditorPanel.SCOPE, ('r',), ('redo-change',)),
            ]
        self.cancel_input = [Key.ESCAPE]
        self.key_sequence = []
        self.current_panel = None
        self.previous_focus = None
        self.panels = []
        self.key_help_panel = None
        self.command_input = None
        self.blocking_task = None

        conf = configparser.ConfigParser(interpolation=None)
        conf.read_dict({ALL_SCOPE: {PREF_OPENER: DEFAULT_OPENER,
                                    PREF_SIDECAR: DEFAULT_SIDECARS}})
        self.configuration = metaindex.configuration.BaseConfiguration(conf)
        utils.parse_ls_colors()
        utils.parse_ls_icons()

        self.layout = layouts.get_layout('horizontal')
        assert self.layout is not None
        self.layout = self.layout(self)

        self.named_clipboard = {}
        # bookmarks are tuples (panel type, base path, selected item)
        self.bookmarks = {}
        self.restore_bookmarks()

        # during command execution paints are not executed, but enqueued
        self.prevent_paint = False
        self.queued_paint = None
        self.queued_config_change = None
        # tuples (panel, function: bool) to call upon next start of the main loop
        self.callbacks = queue.Queue()

        self.key_timeout = 100

        self.loading_config = False
        self.load_config_stack = []

        self.select_file_mode = False
        self.select_file_output = None
        self.selected_file = None
        if args.select_file_mode:
            self.select_file_mode = True
            self.select_file_output = args.select_file_output
            self.keys += [
                (DocPanel.SCOPE, (Key.RETURN,), ('select-and-exit',)),
                (FilePanel.SCOPE, (Key.RETURN,), ('select-and-exit',)),
            ]
        self.start_locations = args.location

        self.command_history = []
        if HISTORYFILE.is_file():
            self.command_history = [line for line in HISTORYFILE.read_text().split("\n")
                                         if len(line.strip()) > 0]

        self.metaindexconf = metaindex.configuration.load(args.metaindex_config)
        self.cache = MemoryCache(self.metaindexconf)

        configpath = pathlib.Path(args.config).expanduser().resolve()
        self.load_config_file(configpath)

    def main(self):
        self.cache.start()
        self.command_input = None
        if len(self.start_locations) > 0:
            for location in self.start_locations:
                path = pathlib.Path(location).expanduser().resolve()
                if path.is_dir() and len(location) > 0:
                    self.panels.append(FilePanel(self, path=path))
                else:
                    self.panels.append(DocPanel(self, searchterm=location))
            self.current_panel = self.panels[-1]
        else:
            self.current_panel = DocPanel(self)
            self.panels.append(self.current_panel)
        self.resize_panels()
        self.paint()

        timed_out = False
        redraw = False
        self.screen.timeout(self.key_timeout)

        # quit when there are no panels (left) to display
        while len(self.panels) > 0:
            while not self.callbacks.empty():
                panel, fnc = self.callbacks.get()

                if panel is self.blocking_task:
                    redraw = fnc()

                elif panel in self.panels and fnc() and self.layout.is_visible(panel):
                    try:
                        panel.paint(True)
                    except curses.error:
                        pass
                    redraw = True

            if not timed_out or redraw:
                if self.key_help_panel is not None:
                    self.key_help_panel.paint()

                if self.blocking_task is not None:
                    self.blocking_task.focus()
                elif self.key_help_panel is not None:
                    self.key_help_panel.focus()
                elif self.current_panel is not None:
                    self.current_panel.focus()
                curses.doupdate()

            timed_out = False
            redraw = False

            key = self.read_key()

            if key == Key.TIMEOUT:
                timed_out = True
            
            elif key == Key.RESIZE:
                self.resize_panels()
                self.paint()
            
            elif self.blocking_task is not None:
                self.blocking_task.handle_key(key)

            elif isinstance(self.current_panel, InputLine):
                self.current_panel.handle_key(key)

            elif key in self.cancel_input and len(self.key_sequence) > 0:
                self.key_sequence = []
                self.hide_key_help()

            else:
                self.key_sequence.append(str(key))
                seq = tuple(self.key_sequence)
                resolved = self.resolve_key_sequence()
                if resolved is not None:
                    self.key_sequence = []
                    if len(resolved) == 2:
                        commandname, _ = resolved
                    elif len(resolved) == 1:
                        commandname = resolved[0]
                    else:
                        logger.error(f"Command sequence {seq} has an invalid command tuple {resolved}")
                        continue

                    self.hide_key_help()

                    if commandname in LIST_SCROLL_COMMANDS:
                        if self.current_panel is not None:
                            self.current_panel.handle_key(seq)
                    elif commandname.startswith(':') and not commandname.startswith('::'):
                        prefill = shlex.split(commandname[1:])
                        prefill = shlex.join([prefill[0]] + sum([expand_part(self, part) for part in prefill[1:]], start=[]))
                        self.execute_command("enter-command", prefill)
                    else:
                        args = None
                        if commandname.startswith('::'):
                            args = shlex.split(commandname[2:])
                            commandname = args.pop(0)

                        command = resolve_command(commandname)
                        if command is not None and isinstance(self.current_panel, command.ACCEPT_IN):
                            if args is None:
                                self.execute_command(commandname)
                            else:
                                self.execute_command(commandname, *args)
                else:
                    self.update_key_help()

                    if self.key_help_panel is None:
                        # last resort: maybe a panel has some use for this special key
                        self.current_panel.handle_key(seq)

    def run(self):
        stored_exc = None
        try:
            retval = super().run()
        except Exception as exc:
            stored_exc = exc

        self.cache.quit()

        try:
            HISTORYFILE.write_text("\n".join([command for command in self.command_history
                                                      if len(command.strip()) > 0]))
        except Exception as exc:
            logger.error(f"Could not write to the history file {HISTORYFILE}: {exc}")

        if stored_exc is not None:
            raise stored_exc

        if self.select_file_mode:
            if self.selected_file is None:
                return -1

            fn = self.select_file_output
            if fn == '-':
                fh = sys.stdout
            else:
                fh = open(self.select_file_output, "wt", encoding="utf-8")
            fh.write(str(self.selected_file))
            fh.close()

        return retval

    def execute_command(self, command, *args, **kwargs):
        """Accepts either a Command instance, a Command class or a name of a command"""
        if isinstance(command, str):
            command = resolve_command(command)

        if type(command).__name__ == 'type' and issubclass(command, Command):
            command = command()

        if command is None:
            logger.info("None command executed. Probably programmer problem.")
            return

        cleanup = False

        if not self.prevent_paint:
            self.prevent_paint = True
            self.queued_paint = []
            cleanup = True
        
        command.execute(self.make_context(), *args, **kwargs)

        if cleanup:
            self.prevent_paint = False
            if len(self.queued_paint) > 0:
                self.paint(any(self.queued_paint))
            self.queued_paint = None

    def resolve_key_sequence(self):
        tplkeys = tuple(self.key_sequence)
        for scope, keys, commands in self.keys:
            if scope not in [ANY_SCOPE, getattr(self.current_panel, 'SCOPE', None)]:
                continue
            if keys == tplkeys:
                return commands
        return None

    def update_key_help(self):
        if len(self.key_sequence) == 0:
            return
        curseq = self.key_sequence
        candidates = []
        for scope, keys, cmd in self.keys:
            if scope not in [ANY_SCOPE, getattr(self.current_panel, 'SCOPE', None)]:
                continue
            if list(keys)[:len(curseq)] != curseq:
                continue
            cmdname = cmd[0]
            if cmdname.startswith(':') and not cmdname.startswith('::'):
                candidates.append((keys, cmd))
                continue
            if cmdname.startswith('::'):
                cmdname = shlex.split(cmdname)[0][2:]
            logger.debug('resolving %s', cmdname)
            command = resolve_command(cmdname)
            if command in LIST_SCROLL_COMMANDS:
                candidates.append((keys, cmd))
                continue
            if command is None:
                logger.debug(f"Invalid command in keys: {cmdname}")
                continue
            if not isinstance(self.current_panel, command.ACCEPT_IN):
                logger.debug(f"Inconsistent ACCEPT_IN / scope for command {cmd[0]}")
                continue
            candidates.append((keys, cmd))

        lines = []
        prefix = self.key_sequence
        for item in candidates:
            keys, item = item
            doc = ""
            if len(item) == 1:
                fnc = item[0]
            else:
                fnc, doc = item
            cmd = resolve_command(fnc)
            doc = cmd.__doc__.split("\n")[0] if cmd is not None else doc
            lines.append(["".join(keys[len(prefix):]), fnc, doc])

        if len(candidates) == 0:
            self.key_sequence = []
            self.hide_key_help()
        elif self.key_help_panel is None:
            self.key_help_panel = KeyHelpPanel(lines, self)
            self.key_help_panel.autosize()
            self.paint()
        else:
            self.key_help_panel.candidates = lines
            self.key_help_panel.autosize()
            self.paint()

    def hide_key_help(self):
        if self.key_help_panel is not None:
            self.key_help_panel.destroy()
            self.key_help_panel = None
            self.paint()

    def resize_panels(self):
        if len(self.panels) < 1:
            return

        maxheight, maxwidth = self.size()
        maxheight -= 2

        self.layout.resize_panels()

        if self.command_input is not None:
            self.command_input.resize(1, maxwidth)
            self.command_input.move(maxheight+1, 0)

        if self.key_help_panel is not None:
            self.key_help_panel.autosize()

        if self.blocking_task is not None:
            self.blocking_task.autoresize()

    def refresh(self, force=False):
        logger.debug("global refresh called")
        if len(self.panels) < 1:
            return
        
        self.layout.refresh(force)

        if self.command_input is not None:
            self.command_input.refresh(force)
        if self.key_help_panel is not None:
            self.key_help_panel.refresh(force)
        if self.blocking_task is not None:
            self.blocking_task.refresh(force)
        super().refresh(force)

    def add_panel(self, panel):
        """Add this panel to the list of active panels"""
        self.panels.append(panel)
        self.resize_panels()
        self.paint()

    def activate_panel(self, panel):
        """Activate this panel.
        """
        assert isinstance(panel, Panel)
        if panel in self.panels:
            logger.debug(f"activating panel {panel} #{self.panels.index(panel)}")
        else:
            logger.debug(f"activating panel {panel}")

        if self.current_panel in self.panels:
            self.previous_focus = self.current_panel
        elif self.previous_focus not in self.panels:
            self.previous_focus = None if len(self.panels) == 0 else self.panels[0]
        self.current_panel = panel

        self.layout.activated_panel(panel)

    def close_panel(self, panel):
        if panel not in self.panels:
            return
        panel.on_close()
        idx = self.panels.index(panel)
        refocus = panel is self.current_panel
        self.panels = self.panels[:idx] + self.panels[idx+1:]
        if refocus:
            self.current_panel = None
            if len(self.panels) > 0:
                self.activate_panel(self.panels[max(0, idx-1)])
        self.resize_panels()
        self.paint()

    def paint(self, clear=False):
        if self.screen is None:
            return

        if self.prevent_paint:
            if self.queued_paint is None:
                self.queued_paint = [clear]
            else:
                self.queued_paint.append(clear)
            return
        logger.debug("global paint called")
        if clear:
            height, width = self.size()
            # don't erase the full screen, panels will do that
            # only clear the lines that are outside the regular panels
            try:
                self.screen.addstr(0, 0, " "*width)
                self.screen.addstr(0, height-1, " "*width)
            except curses.error:
                pass
        self.screen.noutrefresh()

        try:
            self.layout.paint(clear)
        except curses.error:
            pass

        try:
            self.paint_focus_bar()
        except curses.error:
            pass

        try:
            if self.blocking_task is not None:
                self.blocking_task.paint()
            if self.command_input is not None:
                self.command_input.paint()
            if self.current_panel is not None and self.current_panel not in self.panels:
                self.current_panel.paint()
        except curses.error:
            pass

    def paint_focus_bar(self):
        _, width = self.size()
        self.screen.addstr(0, 0, " "*width)
        panels = [f' {nr+1}' for nr in range(len(self.panels))]

        curidx = -1
        if self.current_panel in self.panels:
            curidx = self.panels.index(self.current_panel)
        elif self.previous_focus in self.panels:
            curidx = self.panels.index(self.previous_focus)
        logger.debug(f"focus bar #{curidx} ({panels})")

        title = ""
        if 0 <= curidx < len(self.panels):
            activepanel = self.panels[curidx]
            title = activepanel.title()
        self.screen.addstr(0, 0, title[:width-1])
        self.screen.noutrefresh()
        
        x = width - len(''.join(panels))
        for idx, text in enumerate(panels):
            attr = curses.A_STANDOUT if idx == curidx else curses.A_NORMAL
            self.screen.addstr(0, x, text, attr)
            x += len(text)
        self.screen.noutrefresh()

    def set_clipboard(self, content, clipboard=None):
        """Set the content of the clipboard with the given identifier

        A named clipboard is always a flat list of things (usually pathlib.Path items).
        """
        if clipboard is None:
            clipboard = self.DEFAULT_CLIPBOARD

        if clipboard not in self.named_clipboard:
            self.named_clipboard[clipboard] = []

        if content is None:
            self.clear_clipboard(clipboard)
            return
        elif isinstance(content, (list, tuple, set)):
            self.named_clipboard[clipboard] = list(content)
        else:
            self.named_clipboard[clipboard] = [content]
        logger.debug(f"Clipboard '{clipboard}' is now: {self.named_clipboard[clipboard]}")

    def append_to_clipboard(self, content, clipboard=None):
        """Append the content to the clipboard with the given identifier

        See set_clipboard for more details
        """
        if clipboard is None:
            clipboard = self.DEFAULT_CLIPBOARD

        if clipboard not in self.named_clipboard:
            self.named_clipboard[clipboard] = []

        if isinstance(content, (list, tuple, set)):
            self.named_clipboard[clipboard] += list(content)
        else:
            self.named_clipboard[clipboard].append(content)
        logger.debug(f"Clipboard '{clipboard}' is now: {self.named_clipboard[clipboard]}")

    def clear_clipboard(self, clipboard=None):
        """Clear the clipboard with the given identifier

        See set_clipboard for more details
        """
        if clipboard is None:
            clipboard = self.DEFAULT_CLIPBOARD

        if clipboard in self.named_clipboard:
            del self.named_clipboard[clipboard]
        logger.debug(f"Cleared clipboard '{clipboard}'")

    def get_clipboard_content(self, clipboard=None):
        """Get the content of the clipboard with the given identifier

        See set_clipboard for more details
        """
        if clipboard is None:
            clipboard = self.DEFAULT_CLIPBOARD

        if clipboard not in self.named_clipboard:
            return None

        return self.named_clipboard[clipboard]

    def save_bookmark(self, mark, panel, path, item):
        """Store this bookmark"""
        self.bookmarks[mark] = (type(panel), path, item)
        self.store_bookmarks()

    def load_bookmark(self, mark):
        """Restore this bookmark, create a panel if required"""

        if mark not in self.bookmarks:
            self.error("No such bookmark '{mark}'")
            return

        paneltype, path, item = self.bookmarks[mark]

        if type(self.current_panel) is paneltype and (not hasattr(self.current_panel, 'is_busy') or not self.current_panel.is_busy):
            self.current_panel.jump_to(item, path)
        else:
            if issubclass(paneltype, FilePanel):
                panel = paneltype(self, path=path)
            elif issubclass(paneltype, DocPanel):
                panel = paneltype(self, searchterm=path)
            else:
                msg = f"Unexpected panel type in bookmarks: {paneltype}"
                logger.error(msg)
                self.error(msg)
                return
            self.add_panel(panel)
            self.activate_panel(panel)
            
            panel.jump_to(item)

    def restore_bookmarks(self):
        """Load the previously defined bookmarks from disk

        Bookmarks are stored one per line in plain text in this format:

            <mark> [f|d] <path> <path to item>
        
        """
        if not BOOKMARKFILE.is_file():
            return

        for line in BOOKMARKFILE.read_text().split("\n"):
            parts = shlex.split(line)
            if len(parts) != 4:
                continue
            if parts[1] not in 'fd':
                continue

            mark, typename, path, item = parts
            if len(item) == 0:
                item = None
            if item is not None:
                item = pathlib.Path(item)

            paneltype = None
            if typename == 'f':
                paneltype = FilePanel
                path = pathlib.Path(path)
            elif typename == 'd':
                paneltype = DocPanel
            else:
                raise NotImplementedError(f"Unknown panel type {typename}")

            self.bookmarks[mark] = (paneltype, path, item)

    def store_bookmarks(self):
        """Save the bookmarks to disk"""
        with open(BOOKMARKFILE, 'wt', encoding='utf-8') as fh:
            for mark in sorted(self.bookmarks.keys()):
                typename = 'd' if self.bookmarks[mark][0] is DocPanel else 'f'
                line = [mark, typename] + [str(v) for v in self.bookmarks[mark][1:]]
                fh.write(shlex.join(line) + "\n")

    def configuration_changed(self, change):
        if self.queued_config_change is not None:
            self.queued_config_change.append(change)
        else:
            self.queued_config_change = [change]
            self.apply_config_changes()

    def apply_config_changes(self):
        if len(self.queued_config_change) > 0:
            for scope, name in self.queued_config_change:
                for panel in self.panels:
                    if scope in [panel.SCOPE, 'all']:
                        panel.configuration_changed(name)
        self.queued_config_change = None

    def load_config_file(self, path, context=None):
        path = path.resolve()

        if not path.is_file():
            return

        if path in self.load_config_stack:
            return

        if context is None:
            context = Context(application=self, panel=None)

        if self.queued_config_change is None:
            self.queued_config_change = []

        # prevent duplicate loading of this file
        self.loading_config = True
        self.load_config_stack.append(path)

        logger.info(f"Loading configuration file {path}")

        with open(path, 'rt', encoding='utf-8') as fh:
            for linenr, line in enumerate(fh.readlines()):
                tokens = shlex.split(line.strip())
                if len(tokens) == 0 or tokens[0] == '#':
                    continue

                cmdname = tokens[0]
                args = tokens[1:]

                cmd = resolve_command(cmdname)
                if cmd is None:
                    logger.error(f"Error in configuration file {path}, line {linenr+1}: command '{cmdname}' not found")
                    continue

                cmd().execute(context, *args)

        self.load_config_stack.pop()
        self.loading_config = len(self.load_config_stack) > 0
        if not self.loading_config:
            self.apply_config_changes()

    def open_file(self, path):
        assert path.is_file()
        opener = shlex.split(self.configuration.get("all", "opener", DEFAULT_OPENER))
        if len(opener) == 0:
            # TODO: check for existing file
            return
        with ShellContext(self.screen):
            subprocess.run(opener + [str(path)])
        self.paint(True)

    def get_text_editor(self, show_error=False):
        """Resolve the external text editor to use"""
        editor = [self.configuration.get('all', 'editor', None)] \
               + [os.getenv(name) for name in ['VISUAL', 'EDITOR']]
        for value in editor[:]:
            if value is None:
                continue
            editor = shutil.which(value)
            if editor is not None:
                return shlex.split(editor)

        if show_error:
            msg = "Don't know what editor to use. Please set the 'editor' configuration option"
            logger.error(msg)
            self.error(msg)
        return None

    def make_command_input(self, text=""):
        assert self.command_input is None
        self.command_input = CommandInput(self, 1, (0, 0), prefix=":", text=text)
        self.command_input.cursor = len(text)
        self.resize_panels()
        return self.command_input

    def make_context(self):
        return Context(application=self,
                       panel=self.current_panel)

    def confirm(self, text, default, callback):
        height, width = self.size()
        yes = 'Y' if default == 'y' else 'y'
        no = 'N' if default == 'n' else 'n'
        options = f' ({yes}/{no}) '
        text = text[:width - len(options) - 1] + options
        panel = YesNoConfirmation(callback, default, self, width, (height-1, 0), prefix=text)
        self.activate_panel(panel)
        self.paint()
        panel.paint()

    def error(self, text):
        logger.debug(f"Logging error: {text}")
        if self.screen is not None:
            self.info(text, colors.attr(colors.RED))

    def info(self, text, attr=0):
        if self.screen is None:
            return

        size = self.size()
        display = text[:size[1]] + " "*max(0, size[1]-len(text))
        try:
            self.screen.addstr(size[0]-1, 0, display, attr)
        except:
            pass
        self.screen.refresh()

    def humanize(self, value):
        """Convert any metadata value to a human-friendly string"""
        if isinstance(value, datetime.datetime):
            value = value.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(value, datetime.date):
            value = value.strftime("%Y-%m-%d")
        elif isinstance(value, (int, float, bool)):
            value = str(value)

        assert isinstance(value, str)

        return value.replace('\0', '').strip()

    def get_preferred_sidecar_stores(self):
        """Returns a list of all available metadata stores, sorted by preference of the user"""
        preferred = self.configuration.list(ALL_SCOPE, PREF_SIDECAR, DEFAULT_SIDECARS)
        preferred = [stores.BY_SUFFIX[suffix] for suffix in preferred if suffix in stores.BY_SUFFIX]
        return preferred \
             + [store for store in stores.STORES if store not in preferred]

    def get_editable_sidecar_file(self, filepath):
        """Get the editable sidecar file path for this file

        Returns a tuple (path, is_collection, store) with the pathlib.Path to the sidecar file
        (which may or may not exist), a boolean whether or not the sidecar file is a collection
        file, and the store module that can be used to read/write the metadata to this sidecar.

        May return (None, False, None) in case there is no usable storage. In that case it will
        write a message into the error log, too.
        """
        # find what type of metadata file should be used
        sidecar = None
        is_collection = None
        all_stores = [(store, hasattr(store, 'store')) for store in self.get_preferred_sidecar_stores()]
        usable_stores = [store for store, usable in all_stores if usable]
        if len(usable_stores) == 0:
            logger.error(f"No usable sidecar storage formats available")
            return None, False, None

        prefer_collection = False

        # find existing sidecar file
        for store, usable in all_stores:
            location = filepath.parent / (filepath.stem + store.SUFFIX)
            if location.is_file() and usable:
                sidecar = location
                break

        # if there was none, find existing collection sidecar file
        if sidecar is None:
            for store, is_usable in all_stores:
                for collection_name in self.collection_metadata:
                    location = filepath.parent / (collection_name + store.SUFFIX)
                    if location.is_file():
                        if is_usable:
                            sidecar = location
                            is_collection = True
                            break
                        else:
                            prefer_collection = True
                if sidecar is not None:
                    break
        # still none? just take the first preferred sidecar store and create a sidecar file
        if sidecar is None:
            if prefer_collection:
                sidecar_name = self.collection_metadata[0]
                is_collection = True
            else:
                sidecar_name = filepath.stem
            sidecar = filepath.parent / (sidecar_name + usable_stores[0].SUFFIX)

        return sidecar, is_collection, stores.BY_SUFFIX[sidecar.suffix]

    def determine_sidecar_file(self, filepath):
        """Determines the most likely sidecar file for the given filepath

        Returns tuple (pathlib.Path, collection) with the path to the sidecar file
        and whether or not it is/should be a collection sidecar file.

        CAVEAT: The path to the sidecar file may not exist!
        """
        candidate = None
        preferred_sidecar_format = self.configuration.list(ALL_SCOPE,
                                                           PREF_SIDECAR,
                                                           DEFAULT_SIDECARS)
        all_sidecar_formats = preferred_sidecar_format + [store.SUFFIX for store in stores.STORES]
        # try direct sidecar files first
        for suffix in all_sidecar_formats:
            path = filepath.parent / (filepath.stem + suffix)

            if candidate is None:
                # remember the very first option for later in case we find no existing sidecar file
                candidate = path
            
            if path.is_file():
                return path, False

        # try if there's a collection file
        for suffix in all_sidecar_formats:
            for stem in self.collection_metadata:
                path = filepath.parent / (stem + suffix)
                if path.is_file():
                    return path, True

        # Nothing there, let's just use the first candidate
        return candidate, False

    @property
    def collection_metadata(self):
        return self.configuration.list('General', 'collection-metadata', 'metadata')


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('-m', '--metaindex-config',
                        default=None,
                        type=str,
                        help="The metaindex configuration file to use. Defaults "
                            f"to {metaindex.configuration.CONFIGFILE}.")

    parser.add_argument('-c', '--config',
                        default=CONFIGFILE,
                        type=str,
                        help="Location of the configuration file. Defaults to %(default)s.")

    parser.add_argument('-l', '--log-level',
                        default='warning',
                        choices=['debug', 'info', 'warning', 'error', 'fatal'],
                        help="The level of logging. Defaults to %(default)s.")
    parser.add_argument('--log-file',
                        default=LOGFILE,
                        type=str,
                        help="Where to write the log file to. Defaults to %(default)s.")

    parser.add_argument('--select-file-mode',
                        default=False,
                        action="store_true",
                        help="Run in mode to select a file.")
    parser.add_argument('--select-file-output',
                        default="-",
                        type=str,
                        help="Where to write the selected file to. Defaults to -, i.e. stdout")

    parser.add_argument("location",
                        type=str,
                        nargs='*',
                        default=None,
                        help="Start with these locations. May either be paths in the filesystem "
                             "or metaindex search terms.")

    result = parser.parse_args()

    return result


def run():
    args = parse_args()

    logger.setup(level=args.log_level.upper(), filename=args.log_file)

    logger.debug("Starting metaindexmanager")

    if args.log_level.lower() in ['debug']:
        metaindex.logger.setup(level=logging.DEBUG, filename=args.log_file.parent / 'metaindex.log')
    else:
        metaindex.logger.setup(logging.ERROR, logger.handler)

    if ADDONSPATH.exists():
        prev_sys_path = sys.path.copy()
        sys.path = [str(ADDONSPATH)]
        for item in ADDONSPATH.iterdir():
            if item.is_file() and item.suffix == '.py':
                importlib.import_module(item.stem)
        sys.path = prev_sys_path

    try:
        return Application(args).run()
    except Exception as exc:
        callstack = ''.join(traceback.format_tb(exc.__traceback__))
        logger.fatal(f"The program ended with this exception: {exc}, "
                     f"please report this to the developer\n{callstack}")
        print(f"The program ended with this exception: {exc}. "
               "Please report it to the developers.")
        if args.log_file is not None:
            print(f"Details can be found in the log file at {args.log_file}")
        if args.log_level.lower() == 'debug':
            raise
        return 1


if __name__ == '__main__':
    sys.exit(run())




import traceback
import shlex

from cursedspace import InputLine, Key, Completion

import metaindexmanager.command
from .utils import logger
from .command import resolve_command


class CommandCompletion(Completion):
    def update(self, y, x):
        targetpanel = self.app.previous_focus
        alternatives = []
        text = self.inputline.text[:self.inputline.cursor]
        try:
            parts = shlex.split(text)
        except ValueError:
            return

        if len(parts) <= 1:
            if len(parts) == 0:
                parts = [""]
            alternatives = [name for name, cmd in sorted(metaindexmanager.command._registered_commands.items())
                            if (len(parts[0]) == 0 or name.startswith(parts[0])) and isinstance(targetpanel, cmd.ACCEPT_IN)]

        if len(parts) > 1 or text.endswith(' '):
            cmd = resolve_command(parts[0])
            if cmd is not None and isinstance(targetpanel, cmd.ACCEPT_IN):
                alternatives = cmd().completion_options(self.app.make_context(), *parts[1:])

        if len(alternatives) > 0:
            self.set_alternatives(alternatives, (y, x))
            self.app.paint()
        elif self.is_visible:
            self.close()
            self.app.paint()


class CommandInput(InputLine):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.history_cursor = None
        self.cached_text = None
        self.cached_cursor = None
        self.candidates = []

        self.completion = CommandCompletion(self)

    def handle_key(self, key):
        must_repaint = False

        if key in [Key.TAB, "^N"] and self.completion is not None and not self.completion.is_visible:
            self.update_completion()
        elif self.completion is not None and self.completion.is_visible and self.completion.handle_key(key):
            pass
        elif key in [Key.ESCAPE, "^C"]:
            self.app.execute_command('cancel-command')
        
        elif key in [Key.BACKSPACE] and (self.text is None or len(self.text) == 0):
            self.app.execute_command('cancel-command')
        
        elif key in [Key.UP]:
            if self.history_cursor is None:
                self.cached_text = self.text
                self.cached_cursor = self.cursor
                self.candidates = [line for line in self.app.command_history
                                        if line.startswith(self.cached_text[:self.cursor])]
            if len(self.candidates) > 0:
                if self.history_cursor is None:
                    self.history_cursor = len(self.candidates) - 1
                elif self.history_cursor > 0:
                    self.history_cursor -= 1
                self.text = self.candidates[self.history_cursor]
                self.cursor = min(len(self.text), self.cursor)
                must_repaint = True
        
        elif key in [Key.DOWN] and self.history_cursor is not None:
            if self.history_cursor == len(self.candidates) - 1:
                self.text = self.cached_text
                self.cursor = self.cached_cursor
                self.history_cursor = None
            else:
                self.history_cursor += 1
                self.text = self.candidates[self.history_cursor]
                self.cursor = min(len(self.text), self.cursor)
            must_repaint = True

        elif key in [Key.RETURN]:
            self.app.execute_command('cancel-command')
            seq = shlex.split(self.text)
            if len(seq) == 0:
                command = None
            else:
                command = resolve_command(seq[0])

            if command is not None:
                if not isinstance(self.app.current_panel, command.ACCEPT_IN):
                    error = f"Command {seq[0]} not valid in this panel"
                    logger.error(error)
                    self.app.error(error)
                else:
                    try:
                        if len(seq) > 1:
                            expanded = sum([expand_part(self.app, part) for part in seq[1:]], start=[])
                            self.app.execute_command(command, *expanded)
                        else:
                            self.app.execute_command(command)
                        self.app.command_history.append(self.text)
                        self.text = ""
                    except Exception as exc:
                        error = f"Failed to execute {seq[0]}: {exc}"
                        self.app.error(error)
                        logger.error(error)
                        logger.debug(''.join(traceback.format_tb(exc.__traceback__)))
            elif len(seq) > 0:
                error = f"No such command: {seq[0]}"
                logger.warning(error)
                self.app.error(error)
            self.history_cursor = None
            self.cached_text = None
            self.cached_cursor = None
        
        else:
            text = self.text
            super().handle_key(key)

            if text != self.text and self.history_cursor is not None:
                self.history_cursor = None
                self.cached_text = None
                self.cached_cursor = None

            elif text != self.text and self.completion is not None:
                self.update_completion()

        if must_repaint:
            self.paint()
            self.focus()

    def destroy(self):
        if self.win is not None:
            self.win.erase()
            self.win.refresh()
        super().destroy()


def expand_part(app, part):
    """Expand the single string 'part' into a list of strings

    Expands:
     - '%n' to [filename of the selected item]
     - '%f' to [full path of selected item]
     - '%s' to [full path of selected item 1, full path of selected item 2, ...]
     - '%p' to [full path of folder of selected item]
    """
    if part == '%n':
        return [str(app.current_panel.selected_path.name)]
    elif part == '%f':
        return [str(app.current_panel.selected_path)]
    elif part == '%s':
        return [str(p) for p in app.current_panel.selected_paths]
    elif part == '%p':
        return [str(app.current_panel.selected_path.parent)]
    return [part]




"""Layout definitions"""
from cursedspace import Panel

from metaindexmanager.utils import logger


_registered_layouts = {}


def registered_layout(cls):
    assert issubclass(cls, Layout)
    global _registered_layouts
    assert cls.NAME not in _registered_layouts
    _registered_layouts[cls.NAME] = cls
    return cls


def layouts():
    return list(_registered_layouts.keys())


def get_layout(name):
    """Return the layout by name or None if it doesn't exist"""
    return _registered_layouts.get(name, None)


class Layout:
    def __init__(self, application):
        self.app = application

    def refresh(self, force):
        """Called during Application.refresh"""
        raise NotImplementedError()

    def paint(self, clear=False):
        """Paint all panels on the screen"""
        raise NotImplementedError()

    def resize_panels(self):
        """Resize all panels to fit on the screen"""
        raise NotImplementedError()

    def activated_panel(self, panel):
        """Called after 'panel' has been activated"""
        self.app.paint_focus_bar()

    def is_visible(self, panel):
        """Whether on not this panel is visible"""
        raise NotImplementedError()


@registered_layout
class HorizontalLayout(Layout):
    NAME = 'horizontal'

    def refresh(self, force):
        for panel in self.app.panels:
            panel.refresh(force)

    def resize_panels(self):
        maxheight, maxwidth = self.app.size()
        maxheight -= 2  # border
        posx = 0

        width = maxwidth // len(self.app.panels)
        for panel in self.app.panels:
            panelwidth = width
            if panel is self.app.panels[-1]:
                panelwidth = maxwidth - posx
            panel.border = Panel.BORDER_ALL
            panel.resize(maxheight, panelwidth)
            panel.move(1, posx)
            logger.debug(f"Resized panel {panel} to {maxheight},{panelwidth}")
            posx += width

    def paint(self, clear=False):
        for panel in self.app.panels:
            panel.paint(clear)
            panel.win.refresh()

    def is_visible(self, panel):
        return True


@registered_layout
class TabbedLayout(Layout):
    NAME = 'tabbed'

    def active_panel(self):
        if self.app.current_panel in self.app.panels:
            return self.app.current_panel
        elif self.app.previous_focus in self.app.panels:
            return self.app.previous_focus

        if len(self.app.panels) > 0:
            logger.error(f"Tabbed layout could not determine active panel: current:{self.app.current_panel} prev_focus:{self.app.previous_focus}")
            return self.app.panels[0]
        return None

    def is_visible(self, panel):
        return panel is self.active_panel()

    def refresh(self, force):
        active = self.active_panel()

        if active is not None:
            active.refresh(force)

    def resize_panels(self):
        maxheight, maxwidth = self.app.size()
        maxheight -= 2  # border

        for panel in self.app.panels:
            panel.border = Panel.BORDER_ALL
            panel.resize(maxheight, maxwidth)
            panel.move(1, 0)

    def activated_panel(self, panel):
        self.app.paint(True)

    def paint(self, clear=False):
        activepanel = self.active_panel()

        if activepanel is not None:
            activepanel.paint(clear)
            activepanel.win.refresh()




from abc import ABC

from cursedspace import Panel


_registered_commands = {}


def registered_command(cls):
    assert issubclass(cls, Command)
    global _registered_commands
    assert cls.NAME not in _registered_commands
    _registered_commands[cls.NAME] = cls
    return cls


def resolve_command(name):
    global _registered_commands
    return _registered_commands.get(name, None)


class Command:
    ACCEPT_IN = (Panel,)

    def execute(self, context):
        raise NotImplementedError()

    def completion_options(self, context, *args):
        """Return all possible alternatives for the last argument of *args"""
        return []


def simple_command(name, accept_in=None):
    if accept_in is None:
        accept_in = (Panel,)

    def register_simple_command(fnc):
        docstr = fnc.__doc__
        if docstr is None or len(docstr) == 0:
            docstr = name
        cls = type('SimpleCommand', (Command,), {'__doc__': docstr})
        cls.NAME = name
        cls.ACCEPT_IN = accept_in
        cls.execute = lambda self, context, *args, **kwargs: fnc(context, *args, **kwargs)
        return registered_command(cls)
    return register_simple_command

